import io
import json
import os
import tempfile

from nodus.lsp.server import LanguageServer


def _encode_message(payload: dict) -> bytes:
    body = json.dumps(payload).encode("utf-8")
    return f"Content-Length: {len(body)}\r\n\r\n".encode("ascii") + body


def _decode_messages(data: bytes) -> list[dict]:
    out: list[dict] = []
    stream = io.BytesIO(data)
    while True:
        header = stream.readline()
        if not header:
            break
        if not header.strip():
            continue
        length = int(header.decode("ascii").split(":", 1)[1].strip())
        blank = stream.readline()
        if blank not in {b"\r\n", b"\n"}:
            raise AssertionError("Invalid LSP framing")
        body = stream.read(length)
        out.append(json.loads(body.decode("utf-8")))
    return out


def _run_session(messages: list[dict]) -> list[dict]:
    input_bytes = b"".join(_encode_message(message) for message in messages)
    output = io.BytesIO()
    server = LanguageServer(io.BytesIO(input_bytes), output)
    server.run()
    return _decode_messages(output.getvalue())


def _diagnostic_notifications(messages: list[dict]) -> list[dict]:
    return [message for message in messages if message.get("method") == "textDocument/publishDiagnostics"]


def _file_uri(path: str) -> str:
    return f"file:///{path.replace(os.sep, '/')}"


def test_syntax_errors_are_reported_with_locations():
    with tempfile.TemporaryDirectory() as td:
        path = os.path.join(td, "bad.nd")
        uri = _file_uri(path)
        messages = _run_session(
            [
                {"jsonrpc": "2.0", "id": 1, "method": "initialize", "params": {}},
                {
                    "jsonrpc": "2.0",
                    "method": "textDocument/didOpen",
                    "params": {"textDocument": {"uri": uri, "languageId": "nodus", "version": 1, "text": "let x =\n"}},
                },
                {"jsonrpc": "2.0", "id": 2, "method": "shutdown", "params": {}},
                {"jsonrpc": "2.0", "method": "exit", "params": {}},
            ]
        )
    diagnostics = _diagnostic_notifications(messages)
    assert len(diagnostics) == 1
    payload = diagnostics[0]["params"]["diagnostics"][0]
    assert diagnostics[0]["params"]["uri"] == uri
    assert payload["severity"] == 1
    assert payload["source"] == "nodus"
    assert payload["file"].endswith("bad.nd")
    assert payload["line"] == 1
    assert payload["column"] is not None


def test_cross_module_errors_publish_for_imported_file():
    with tempfile.TemporaryDirectory() as td:
        main_path = os.path.join(td, "main.nd")
        lib_path = os.path.join(td, "lib.nd")
        with open(lib_path, "w", encoding="utf-8") as handle:
            handle.write("export fn broken( {\n")
        uri = _file_uri(main_path)
        messages = _run_session(
            [
                {"jsonrpc": "2.0", "id": 1, "method": "initialize", "params": {}},
                {
                    "jsonrpc": "2.0",
                    "method": "textDocument/didOpen",
                    "params": {
                        "textDocument": {
                            "uri": uri,
                            "languageId": "nodus",
                            "version": 1,
                            "text": 'import "./lib.nd" as lib\nprint(1)\n',
                        }
                    },
                },
                {"jsonrpc": "2.0", "id": 2, "method": "shutdown", "params": {}},
                {"jsonrpc": "2.0", "method": "exit", "params": {}},
            ]
        )
    notifications = _diagnostic_notifications(messages)
    lib_notification = next(item for item in notifications if item["params"]["uri"].endswith("/lib.nd"))
    assert lib_notification["params"]["diagnostics"]
    assert "Expected ID" in lib_notification["params"]["diagnostics"][0]["message"]


def test_diagnostics_clear_after_fix():
    with tempfile.TemporaryDirectory() as td:
        path = os.path.join(td, "main.nd")
        uri = _file_uri(path)
        messages = _run_session(
            [
                {"jsonrpc": "2.0", "id": 1, "method": "initialize", "params": {}},
                {
                    "jsonrpc": "2.0",
                    "method": "textDocument/didOpen",
                    "params": {"textDocument": {"uri": uri, "languageId": "nodus", "version": 1, "text": "let x =\n"}},
                },
                {
                    "jsonrpc": "2.0",
                    "method": "textDocument/didChange",
                    "params": {
                        "textDocument": {"uri": uri, "version": 2},
                        "contentChanges": [{"text": "let x = 1\nprint(x)\n"}],
                    },
                },
                {"jsonrpc": "2.0", "id": 2, "method": "shutdown", "params": {}},
                {"jsonrpc": "2.0", "method": "exit", "params": {}},
            ]
        )
    notifications = [item for item in _diagnostic_notifications(messages) if item["params"]["uri"] == uri]
    assert notifications[0]["params"]["diagnostics"]
    assert notifications[-1]["params"]["diagnostics"] == []


def test_incremental_diagnostics_update_dependents():
    with tempfile.TemporaryDirectory() as td:
        main_path = os.path.join(td, "main.nd")
        util_path = os.path.join(td, "util.nd")
        main_uri = _file_uri(main_path)
        util_uri = _file_uri(util_path)
        with open(main_path, "w", encoding="utf-8") as handle:
            handle.write('import "./util.nd" as util\nprint(util.value)\n')
        with open(util_path, "w", encoding="utf-8") as handle:
            handle.write("export let value = 1\n")
        messages = _run_session(
            [
                {"jsonrpc": "2.0", "id": 1, "method": "initialize", "params": {}},
                {
                    "jsonrpc": "2.0",
                    "method": "textDocument/didOpen",
                    "params": {
                        "textDocument": {
                            "uri": main_uri,
                            "languageId": "nodus",
                            "version": 1,
                            "text": 'import "./util.nd" as util\nprint(util.value)\n',
                        }
                    },
                },
                {
                    "jsonrpc": "2.0",
                    "method": "textDocument/didOpen",
                    "params": {
                        "textDocument": {
                            "uri": util_uri,
                            "languageId": "nodus",
                            "version": 1,
                            "text": "export let value = 1\n",
                        }
                    },
                },
                {
                    "jsonrpc": "2.0",
                    "method": "textDocument/didChange",
                    "params": {
                        "textDocument": {"uri": util_uri, "version": 2},
                        "contentChanges": [{"text": "export let other = 1\n"}],
                    },
                },
                {"jsonrpc": "2.0", "id": 2, "method": "shutdown", "params": {}},
                {"jsonrpc": "2.0", "method": "exit", "params": {}},
            ]
        )
    notifications = _diagnostic_notifications(messages)
    main_updates = [item for item in notifications if item["params"]["uri"] == main_uri]
    util_updates = [item for item in notifications if item["params"]["uri"] == util_uri]
    assert main_updates[0]["params"]["diagnostics"] == []
    assert util_updates[0]["params"]["diagnostics"] == []
    dependent_update = next(item for item in reversed(main_updates) if item["params"]["diagnostics"])
    assert "has no member 'value'" in dependent_update["params"]["diagnostics"][0]["message"]



import io
import os
import tempfile
import unittest
from contextlib import redirect_stdout

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader


def run_program(src: str, source_path: str | None = None) -> list[str]:
    vm = lang.VM([], {}, code_locs=[], source_path=source_path)
    _loader = ModuleLoader(project_root=None, vm=vm)
    buf = io.StringIO()
    with redirect_stdout(buf):
        _loader.load_module_from_source(src, module_name=source_path or "<memory>")
    return buf.getvalue().splitlines()


class StdlibTests(unittest.TestCase):
    def test_std_strings(self):
        src = """
import "std:strings" as s
print(s.upper("hello"))
print(s.lower("HELLO"))
print(s.trim("  hi  "))
print(s.split("a,b,c", ","))
print(s.join(["x", "y"], "-"))
print(s.contains("hello", "ell"))
"""
        self.assertEqual(
            run_program(src, source_path="main.nd"),
            ["HELLO", "hello", "hi", "[\"a\", \"b\", \"c\"]", "x-y", "true"],
        )

    def test_std_collections(self):
        src = """
import "std:collections" as c
let nums = [1, 2, 3]
print(c.len(nums))
print(c.map(nums, fn(x) { return x * 2 }))
print(c.filter(nums, fn(x) { return x > 1 }))
print(c.reduce(nums, fn(acc, x) { return acc + x }, 0))
print(c.push(nums, 4))
print(c.pop(nums))
print(nums)
"""
        self.assertEqual(
            run_program(src, source_path="main.nd"),
            ["3.0", "[2.0, 4.0, 6.0]", "[2.0, 3.0]", "6.0", "[1.0, 2.0, 3.0, 4.0]", "4.0", "[1.0, 2.0, 3.0]"],
        )

    def test_std_json(self):
        src = """
import "std:json" as j
let obj = j.parse("{\\"x\\":1,\\"name\\":\\"Alice\\",\\"tags\\":[1,2]}")
print(obj.x)
print(obj.name)
print(j.stringify(obj))
"""
        self.assertEqual(
            run_program(src, source_path="main.nd"),
            ["1.0", "Alice", "{\"x\": 1, \"name\": \"Alice\", \"tags\": [1, 2]}"],
        )

    def test_std_math(self):
        src = """
import "std:math" as m
print(m.abs(-4))
print(m.min(2, 5))
print(m.max(2, 5))
print(m.floor(2.9))
print(m.ceil(2.1))
print(m.sqrt(9))
print(type(m.random()))
"""
        self.assertEqual(
            run_program(src, source_path="main.nd"),
            ["4.0", "2.0", "5.0", "2.0", "3.0", "3.0", "number"],
        )

    def test_std_fs(self):
        with tempfile.TemporaryDirectory() as td:
            td_norm = td.replace("\\", "/")
            file_path = os.path.join(td_norm, "note.txt").replace("\\", "/")
            src = f"""
import "std:fs" as fs
fs.write("{file_path}", "hello")
print(fs.read("{file_path}"))
print(fs.exists("{file_path}"))
print(fs.listdir("{td_norm}"))
"""
            self.assertEqual(
                run_program(src, source_path="main.nd"),
                ["hello", "true", "[\"note.txt\"]"],
            )


if __name__ == "__main__":
    unittest.main()



import io
import json
import os
import task_graph
import tempfile
import time
import unittest
from contextlib import redirect_stderr, redirect_stdout
import http.client

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader
from nodus.services.server import run_in_thread
from nodus.orchestration.workflow_lowering import find_workflow_value


def run_program(src: str, source_path: str = "workflow.nd"):
    _loader = ModuleLoader(project_root=None)
    code, functions, code_locs = _loader.compile_only(
        src,
        module_name=source_path,
    )
    vm = lang.VM(code, functions, code_locs=code_locs, source_path=source_path)
    out_buf = io.StringIO()
    err_buf = io.StringIO()
    with redirect_stdout(out_buf), redirect_stderr(err_buf):
        vm.run()
    return vm, out_buf.getvalue().splitlines(), err_buf.getvalue()


class WorkflowDslTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.server, cls.thread = run_in_thread("127.0.0.1", 0, allowed_paths=["."])
        cls.port = cls.server.server_address[1]
        time.sleep(0.05)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server.server_close()

    def request(self, method: str, path: str, payload: dict | None = None):
        conn = http.client.HTTPConnection("127.0.0.1", self.port, timeout=5)
        body = None
        headers = {}
        if payload is not None:
            body = json.dumps(payload)
            headers["Content-Type"] = "application/json"
        conn.request(method, path, body=body, headers=headers)
        resp = conn.getresponse()
        data = resp.read().decode("utf-8")
        conn.close()
        return resp.status, json.loads(data) if data else {}

    def test_linear_workflow(self):
        src = """
workflow demo {
    step a {
        return 1
    }

    step b after a {
        return a + 1
    }
}

let result = run_workflow(demo)
print(result["steps"]["a"])
print(result["steps"]["b"])
"""
        _vm, out, err = run_program(src)
        self.assertEqual(out, ["1.0", "2.0"])
        self.assertEqual(err.strip(), "")

    def test_workflow_state_initialization(self):
        src = """
workflow demo {
    state x = 1

    step a {
        return x
    }
}

let result = run_workflow(demo)
print(result["steps"]["a"])
print(result["state"]["x"])
"""
        _vm, out, err = run_program(src)
        self.assertEqual(out, ["1.0", "1.0"])
        self.assertEqual(err.strip(), "")

    def test_workflow_state_mutation(self):
        src = """
workflow demo {
    state x = 1

    step a {
        x = x + 1
        return x
    }

    step b after a {
        return x
    }
}

let result = run_workflow(demo)
print(result["steps"]["b"])
print(result["state"]["x"])
"""
        _vm, out, err = run_program(src)
        self.assertEqual(out, ["2.0", "2.0"])
        self.assertEqual(err.strip(), "")

    def test_workflow_checkpoint_recorded(self):
        src = """
workflow demo {
    state x = 1

    step a {
        checkpoint "after_a"
        return x
    }
}

let result = run_workflow(demo)
print(result["checkpoints"][0]["label"])
print(result["checkpoints"][0]["step"])
"""
        _vm, out, err = run_program(src)
        self.assertEqual(out, ["after_a", "a"])
        self.assertEqual(err.strip(), "")

    def test_workflow_checkpoint_rollback(self):
        src = """
workflow demo {
    state x = 0

    step a {
        x = x + 1
        checkpoint "after_a"
        return x
    }

    step b after a {
        return x
    }
}
"""
        _loader = ModuleLoader(project_root=None)
        code, functions, code_locs = _loader.compile_only(src, module_name="rollback.nd")
        vm = lang.VM(code, functions, code_locs=code_locs, source_path="rollback.nd")
        vm.run()
        workflow = find_workflow_value(vm.globals, "demo")
        first = vm.builtin_run_workflow(workflow)
        graph_id = first["graph_id"]
        resumed = vm.builtin_resume_workflow(graph_id, "after_a")
        self.assertEqual(resumed["steps"]["a"], 2)
        self.assertEqual(resumed["steps"]["b"], 2)
        self.assertEqual(resumed["state"]["x"], 2)

    def test_branching_workflow_plan_uses_step_names(self):
        src = """
workflow demo {
    step a { return 1 }
    step b after a { return a + 1 }
    step c after a { return a * 2 }
}

let plan = plan_workflow(demo)
print(len(plan["parallel_groups"][1]))
print(plan["parallel_groups"][1][0] == "b" || plan["parallel_groups"][1][1] == "b")
"""
        _vm, out, _err = run_program(src)
        self.assertEqual(out[0], "2.0")
        self.assertEqual(out[1], "true")

    def test_merge_workflow_injects_dependency_values(self):
        src = """
workflow demo {
    step a { return 1 }
    step b { return 2 }
    step c after a, b {
        return a + b
    }
}

let result = run_workflow(demo)
print(result["steps"]["c"])
"""
        _vm, out, _err = run_program(src)
        self.assertEqual(out, ["3.0"])

    def test_workflow_step_options_use_existing_retries(self):
        src = """
let state = { "count": 0 }

workflow demo {
    step flaky with { retries: 2, retry_delay_ms: 1 } {
        if (state["count"] == 0) {
            state["count"] = 1
            throw "fail"
        }
        return 5
    }
}

let result = run_workflow(demo)
print(result["steps"]["flaky"])
print(result["attempts"]["task_1"])
"""
        _vm, out, _err = run_program(src)
        self.assertEqual(out, ["5.0", "2.0"])

    def test_workflow_resume(self):
        src = """
workflow demo {
    state x = 1
    step a {
        x = x + 1
        return x
    }
    step b after a { return x }
}
"""
        _loader = ModuleLoader(project_root=None)
        code, functions, code_locs = _loader.compile_only(src, module_name="resume.nd")
        vm = lang.VM(code, functions, code_locs=code_locs, source_path="resume.nd")
        vm.run()
        workflow = find_workflow_value(vm.globals, "demo")
        first = vm.builtin_run_workflow(workflow)
        graph_id = first["graph_id"]
        path = os.path.join(".nodus", "graphs", f"{graph_id}.json")
        with open(path, "w", encoding="utf-8") as f:
            json.dump(
                {
                    "graph_id": graph_id,
                    "status": "running",
                    "metadata": {
                        "workflow_name": "demo",
                        "step_to_task": {"a": "task_1", "b": "task_2"},
                        "task_to_step": {"task_1": "a", "task_2": "b"},
                    },
                    "workflow_state": {"x": 2},
                    "checkpoints": [{"label": "after_a", "step": "a", "timestamp": 1, "state": {"x": 2}}],
                    "tasks": {
                        "task_1": {"state": "completed", "result": 2, "attempts": 1, "step_name": "a"},
                        "task_2": {"state": "pending", "attempts": 0, "step_name": "b"},
                    },
                },
                f,
            )
        resumed = vm.builtin_resume_workflow(graph_id, "after_a")
        self.assertEqual(resumed["steps"]["b"], 3)
        self.assertEqual(resumed["state"]["x"], 3)

    def test_workflow_resume_rebuilds_from_persisted_metadata(self):
        src = """
workflow demo {
    state x = 1

    step a {
        x = x + 1
        checkpoint "after_a"
        return x
    }

    step b after a {
        return x
    }
}

let result = run_workflow(demo)
"""
        with tempfile.TemporaryDirectory() as td:
            path = os.path.join(td, "rebuild.nd")
            with open(path, "w", encoding="utf-8") as f:
                f.write(src)
            _loader = ModuleLoader(project_root=None)
            code, functions, code_locs = _loader.compile_only(src, module_name=path)
            vm = lang.VM(code, functions, code_locs=code_locs, source_path=path)
            vm.run()
            result = vm.globals["result"]
            if hasattr(result, "value"):
                result = result.value
            graph_id = result["graph_id"]
            task_graph._GRAPH_REGISTRY.pop(graph_id, None)
            task_graph._GRAPH_VMS.pop(graph_id, None)

            resumed_vm = lang.VM([], {}, code_locs=[], source_path=None)
            resumed = resumed_vm.builtin_resume_workflow(graph_id, "after_a")
            self.assertEqual(resumed["steps"]["a"], 3)
            self.assertEqual(resumed["steps"]["b"], 3)
            self.assertEqual(resumed["state"]["x"], 3)

    def test_workflow_runtime_events(self):
        src = """
workflow demo {
    step a { return 1 }
    step b after a { return a + 1 }
}

run_workflow(demo)
"""
        vm, _out, _err = run_program(src)
        event_types = [event.type for event in vm.event_bus.events()]
        self.assertIn("workflow_start", event_types)
        self.assertIn("workflow_step_start", event_types)
        self.assertIn("workflow_step_complete", event_types)
        self.assertIn("workflow_complete", event_types)

    def test_workflow_api_endpoints(self):
        code = """
workflow demo {
    state x = 1
    step a {
        checkpoint "after_a"
        return x
    }
    step b after a { return a + 1 }
}
"""
        status, payload = self.request("POST", "/workflow/run", {"code": code, "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertEqual(payload["result"]["steps"]["b"], 2)
        self.assertEqual(payload["result"]["state"]["x"], 1)
        self.assertEqual(payload["result"]["checkpoints"][0]["label"], "after_a")
        graph_id = payload["result"]["graph_id"]

        status, payload = self.request("POST", "/workflow/plan", {"code": code, "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertEqual(payload["plan"]["parallel_groups"][1], ["b"])

        status, payload = self.request("GET", f"/workflow/checkpoints/{graph_id}")
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertEqual(payload["checkpoints"][0]["label"], "after_a")

        task_graph._GRAPH_REGISTRY.pop(graph_id, None)
        task_graph._GRAPH_VMS.pop(graph_id, None)
        status, payload = self.request("POST", "/workflow/resume", {"graph_id": graph_id, "checkpoint": "after_a"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertEqual(payload["result"]["state"]["x"], 1)
        self.assertEqual(payload["result"]["checkpoints"][0]["label"], "after_a")
        self.assertEqual(payload["result"]["steps"]["a"], 1)
        self.assertEqual(payload["result"]["steps"]["b"], 2)

    def test_workflow_cli_commands(self):
        code = """
workflow demo {
    state x = 1
    step a {
        checkpoint "after_a"
        return x
    }
    step b after a { return a + 1 }
}
"""
        with tempfile.TemporaryDirectory() as td:
            path = os.path.join(td, "demo.nd")
            with open(path, "w", encoding="utf-8") as f:
                f.write(code)

            out = io.StringIO()
            with redirect_stdout(out):
                run_exit = lang.main(["nodus", "workflow-run", path])
            self.assertEqual(run_exit, 0)
            run_payload = json.loads(out.getvalue().strip())
            self.assertEqual(run_payload["steps"]["b"], 2)
            graph_id = run_payload["graph_id"]

            out = io.StringIO()
            with redirect_stdout(out):
                plan_exit = lang.main(["nodus", "workflow-plan", path])
            self.assertEqual(plan_exit, 0)
            plan_payload = json.loads(out.getvalue().strip())
            self.assertEqual(plan_payload["parallel_groups"][1], ["b"])

            out = io.StringIO()
            with redirect_stdout(out):
                checkpoints_exit = lang.main(["nodus", "workflow-checkpoints", graph_id])
            self.assertEqual(checkpoints_exit, 0)
            checkpoints_payload = json.loads(out.getvalue().strip())
            self.assertEqual(checkpoints_payload[0]["label"], "after_a")

            task_graph._GRAPH_REGISTRY.pop(graph_id, None)
            task_graph._GRAPH_VMS.pop(graph_id, None)
            out = io.StringIO()
            with redirect_stdout(out):
                resume_exit = lang.main(["nodus", "workflow-resume", graph_id, "--checkpoint", "after_a"])
            self.assertEqual(resume_exit, 0)
            resume_payload = json.loads(out.getvalue().strip())
            self.assertEqual(resume_payload["steps"]["b"], 2)


if __name__ == "__main__":
    unittest.main()



import io
import unittest
from contextlib import redirect_stderr, redirect_stdout

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader


def run_program(src: str, source_path: str | None = None) -> tuple[list[str], str]:
    vm = lang.VM([], {}, code_locs=[], source_path=source_path)
    _loader = ModuleLoader(project_root=None, vm=vm)
    out_buf = io.StringIO()
    err_buf = io.StringIO()
    with redirect_stdout(out_buf), redirect_stderr(err_buf):
        _loader.load_module_from_source(src, module_name=source_path or "<memory>")
    return out_buf.getvalue().splitlines(), err_buf.getvalue()


class SchedulerTests(unittest.TestCase):
    def test_interleaves_coroutines(self):
        src = """
fn worker(name) {
    print("start " + name)
    yield
    print("resume " + name)
}

let a = coroutine(fn() { worker("A") })
let b = coroutine(fn() { worker("B") })
spawn(a)
spawn(b)
run_loop()
"""
        out, err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["start A", "start B", "resume A", "resume B"])
        self.assertEqual(err.strip(), "")

    def test_sleep_allows_other_tasks_to_run(self):
        src = """
import "std:async"

fn sleeper() {
    print("wait")
    sleep(50)
    print("done")
}

fn fast() {
    print("fast")
    yield
    print("fast2")
}

let a = coroutine(fn() { sleeper() })
let b = coroutine(fn() { fast() })
spawn(a)
spawn(b)
run_loop()
"""
        out, err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["wait", "fast", "fast2", "done"])
        self.assertEqual(err.strip(), "")

    def test_error_isolation(self):
        src = """
fn bad() {
    throw "boom"
}

fn good() {
    print("ok")
}

spawn(coroutine(bad))
spawn(coroutine(good))
run_loop()
"""
        out, err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["ok"])
        self.assertIn("Runtime error", err)
        self.assertIn("boom", err)


if __name__ == "__main__":
    unittest.main()



import io
import json
import os
from contextlib import redirect_stdout

import nodus as lang

from nodus.runtime.module_loader import ModuleLoader


def _write(tmp_path, name: str, content: str) -> str:
    path = tmp_path / name
    path.write_text(content, encoding="utf-8")
    return str(path)


def _bump_mtime(path: str) -> None:
    current = os.stat(path).st_mtime_ns
    updated = current + 2_000_000_000
    os.utime(path, ns=(updated, updated))


def test_unchanged_modules_skip_reparsing_and_recompilation(tmp_path, monkeypatch):
    (tmp_path / "nodus.toml").write_text('name = "demo"\nversion = "0.1.0"\n', encoding="utf-8")
    util_path = _write(tmp_path, "util.nd", "export let value = 1\n")
    main_path = _write(tmp_path, "main.nd", 'import "./util.nd" as util\nprint(util.value)\n')

    ModuleLoader(project_root=str(tmp_path)).load_module_from_path(main_path)

    parse_calls = []
    compile_calls = []
    original_parse = ModuleLoader._parse_module
    original_compile = ModuleLoader._compile_module

    def counting_parse(self, module_id, **kwargs):
        parse_calls.append(module_id)
        return original_parse(self, module_id, **kwargs)

    def counting_compile(self, metadata):
        compile_calls.append(metadata.module_id)
        return original_compile(self, metadata)

    monkeypatch.setattr(ModuleLoader, "_parse_module", counting_parse)
    monkeypatch.setattr(ModuleLoader, "_compile_module", counting_compile)

    ModuleLoader(project_root=str(tmp_path)).load_module_from_path(main_path)

    assert parse_calls == []
    assert compile_calls == []


def test_dependency_change_triggers_recompilation(tmp_path, monkeypatch):
    (tmp_path / "nodus.toml").write_text('name = "demo"\nversion = "0.1.0"\n', encoding="utf-8")
    util_path = _write(tmp_path, "util.nd", "export let value = 1\n")
    main_path = _write(tmp_path, "main.nd", 'import "./util.nd" as util\nprint(util.value)\n')

    ModuleLoader(project_root=str(tmp_path)).load_module_from_path(main_path)

    with open(util_path, "w", encoding="utf-8") as handle:
        handle.write("export let value = 2\n")
    _bump_mtime(util_path)

    compile_calls = []
    original_compile = ModuleLoader._compile_module

    def counting_compile(self, metadata):
        compile_calls.append(metadata.module_id)
        return original_compile(self, metadata)

    monkeypatch.setattr(ModuleLoader, "_compile_module", counting_compile)

    ModuleLoader(project_root=str(tmp_path)).load_module_from_path(main_path)

    assert set(compile_calls) == {os.path.abspath(main_path), os.path.abspath(util_path)}


def test_dependency_graph_persists_and_cli_prints_it(tmp_path):
    (tmp_path / "nodus.toml").write_text('name = "demo"\nversion = "0.1.0"\n', encoding="utf-8")
    util_path = _write(tmp_path, "util.nd", "export let value = 1\n")
    main_path = _write(tmp_path, "main.nd", 'import "./util.nd" as util\nprint(util.value)\n')

    ModuleLoader(project_root=str(tmp_path)).load_module_from_path(main_path)

    graph_path = tmp_path / ".nodus" / "deps.json"
    assert graph_path.is_file()

    payload = json.loads(graph_path.read_text(encoding="utf-8"))
    modules = payload["modules"]
    assert os.path.abspath(main_path) in modules
    assert os.path.abspath(util_path) in modules
    assert modules[os.path.abspath(main_path)]["imports"] == [os.path.abspath(util_path)]
    assert modules[os.path.abspath(util_path)]["imports"] == []

    stdout = io.StringIO()
    with redirect_stdout(stdout):
        exit_code = lang.main(["nodus", "deps", "--project-root", str(tmp_path)])

    assert exit_code == 0
    cli_payload = json.loads(stdout.getvalue())
    assert cli_payload == payload



"""Source formatter for Nodus."""

from nodus.frontend.ast.ast_nodes import (
    Assign,
    ActionStmt,
    Attr,
    Bin,
    Block,
    Bool,
    Call,
    Comment,
    DestructureLet,
    ExportFrom,
    ExportList,
    ExprStmt,
    FieldAssign,
    FnDef,
    FnExpr,
    For,
    ForEach,
    GoalDef,
    GoalStep,
    If,
    Import,
    Index,
    IndexAssign,
    Let,
    ListLit,
    ListPattern,
    MapLit,
    Nil,
    Num,
    Param,
    Print,
    RecordLiteral,
    RecordPattern,
    Return,
    Str,
    Throw,
    TryCatch,
    Unary,
    Var,
    VarPattern,
    WorkflowDef,
    WorkflowStep,
    WorkflowStateDecl,
    CheckpointStmt,
    While,
    Yield,
)
from nodus.frontend.lexer import tokenize
from nodus.frontend.parser import Parser


INDENT = "    "


def format_source(src: str, keep_trailing_comments: bool = False) -> str:
    stmts = Parser(tokenize(src)).parse()
    return format_program(stmts, keep_trailing_comments=keep_trailing_comments)


def format_program(stmts: list, keep_trailing_comments: bool = False) -> str:
    lines: list[str] = []
    prev_import = False
    prev_fn = False

    for stmt in stmts:
        is_import = isinstance(stmt, Import)
        is_fn = isinstance(stmt, FnDef)
        if lines:
            if prev_import and not is_import:
                lines.append("")
            elif prev_fn or is_fn:
                lines.append("")
        lines.extend(format_stmt(stmt, indent=0, keep_trailing_comments=keep_trailing_comments))
        prev_import = is_import
        prev_fn = is_fn

    return "\n".join(lines).rstrip() + "\n"


def format_stmt(stmt, indent: int, keep_trailing_comments: bool = False) -> list[str]:
    prefix = INDENT * indent
    lines: list[str] = []

    comments = getattr(stmt, "_comments", None)
    if comments:
        for comment in comments:
            lines.append(f"{prefix}{comment.rstrip()}")

    trailing = getattr(stmt, "_trailing_comments", None)

    if isinstance(stmt, Import):
        if stmt.names is not None:
            names = ", ".join(stmt.names)
            lines.append(f"{prefix}import {{ {names} }} from {format_string(stmt.path)}")
            return attach_trailing(lines, prefix, trailing, keep_trailing_comments)
        if stmt.alias is not None:
            lines.append(f"{prefix}import {format_string(stmt.path)} as {stmt.alias}")
            return attach_trailing(lines, prefix, trailing, keep_trailing_comments)
        lines.append(f"{prefix}import {format_string(stmt.path)}")
        return attach_trailing(lines, prefix, trailing, keep_trailing_comments)

    if isinstance(stmt, ExportFrom):
        names = ", ".join(stmt.names)
        lines.append(f"{prefix}export {{ {names} }} from {format_string(stmt.path)}")
        return attach_trailing(lines, prefix, trailing, keep_trailing_comments)

    if isinstance(stmt, ExportList):
        names = ", ".join(stmt.names)
        lines.append(f"{prefix}export {{ {names} }}")
        return attach_trailing(lines, prefix, trailing, keep_trailing_comments)

    if isinstance(stmt, Let):
        name = stmt.name if stmt.type_hint is None else f"{stmt.name}: {stmt.type_hint}"
        if stmt.exported:
            lines.append(f"{prefix}export let {name} = {format_expr(stmt.expr)}")
            return attach_trailing(lines, prefix, trailing, keep_trailing_comments)
        lines.append(f"{prefix}let {name} = {format_expr(stmt.expr)}")
        return attach_trailing(lines, prefix, trailing, keep_trailing_comments)

    if isinstance(stmt, Print):
        lines.append(f"{prefix}print({format_expr(stmt.expr)})")
        return attach_trailing(lines, prefix, trailing, keep_trailing_comments)

    if isinstance(stmt, ExprStmt):
        lines.append(f"{prefix}{format_expr(stmt.expr)}")
        return attach_trailing(lines, prefix, trailing, keep_trailing_comments)

    if isinstance(stmt, Return):
        if stmt.expr is None:
            lines.append(f"{prefix}return")
            return attach_trailing(lines, prefix, trailing, keep_trailing_comments)
        lines.append(f"{prefix}return {format_expr(stmt.expr)}")
        return attach_trailing(lines, prefix, trailing, keep_trailing_comments)

    if isinstance(stmt, FnDef):
        param_text = ", ".join(format_param(param) for param in stmt.params)
        return_text = f" -> {stmt.return_type}" if stmt.return_type else ""
        if stmt.exported:
            header = f"{prefix}export fn {stmt.name}({param_text}){return_text} {{"
        else:
            header = f"{prefix}fn {stmt.name}({param_text}){return_text} {{"
        body_lines = format_block(stmt.body, indent + 1, keep_trailing_comments=keep_trailing_comments)
        return lines + [header] + body_lines + [f"{prefix}}}"] + trailing_lines(prefix, trailing)

    if isinstance(stmt, WorkflowDef):
        header = f"{prefix}workflow {stmt.name} {{"
        body_lines: list[str] = []
        for state in stmt.states:
            body_lines.extend(format_stmt(state, indent + 1, keep_trailing_comments=keep_trailing_comments))
        for step in stmt.steps:
            body_lines.extend(format_stmt(step, indent + 1, keep_trailing_comments=keep_trailing_comments))
        return lines + [header] + body_lines + [f"{prefix}}}"] + trailing_lines(prefix, trailing)

    if isinstance(stmt, GoalDef):
        header = f"{prefix}goal {stmt.name} {{"
        body_lines: list[str] = []
        for state in stmt.states:
            body_lines.extend(format_stmt(state, indent + 1, keep_trailing_comments=keep_trailing_comments))
        for step in stmt.steps:
            body_lines.extend(format_stmt(step, indent + 1, keep_trailing_comments=keep_trailing_comments))
        return lines + [header] + body_lines + [f"{prefix}}}"] + trailing_lines(prefix, trailing)

    if isinstance(stmt, WorkflowStateDecl):
        lines.append(f"{prefix}state {stmt.name} = {format_expr(stmt.value)}")
        return attach_trailing(lines, prefix, trailing, keep_trailing_comments)

    if isinstance(stmt, WorkflowStep):
        deps = ""
        if stmt.deps:
            deps = " after " + ", ".join(stmt.deps)
        options = ""
        if stmt.options is not None:
            options = f" with {format_expr(stmt.options)}"
        header = f"{prefix}step {stmt.name}{deps}{options} {{"
        body_lines = format_block(stmt.body, indent + 1, keep_trailing_comments=keep_trailing_comments)
        return lines + [header] + body_lines + [f"{prefix}}}"] + trailing_lines(prefix, trailing)

    if isinstance(stmt, GoalStep):
        deps = ""
        if stmt.deps:
            deps = " after " + ", ".join(stmt.deps)
        options = ""
        if stmt.options is not None:
            options = f" with {format_expr(stmt.options)}"
        header = f"{prefix}step {stmt.name}{deps}{options} {{"
        body_lines = format_block(stmt.body, indent + 1, keep_trailing_comments=keep_trailing_comments)
        return lines + [header] + body_lines + [f"{prefix}}}"] + trailing_lines(prefix, trailing)

    if isinstance(stmt, If):
        header = f"{prefix}if ({format_expr(stmt.cond)}) {{"
        then_lines = format_block(stmt.then_branch, indent + 1, keep_trailing_comments=keep_trailing_comments)
        out = [header] + then_lines + [f"{prefix}}}"]
        if stmt.else_branch is not None:
            else_header = f"{prefix}else {{"
            else_lines = format_block(stmt.else_branch, indent + 1, keep_trailing_comments=keep_trailing_comments)
            out[-1] = f"{prefix}}} else {{"
            out += else_lines + [f"{prefix}}}"]
        return lines + out + trailing_lines(prefix, trailing)

    if isinstance(stmt, While):
        header = f"{prefix}while ({format_expr(stmt.cond)}) {{"
        body_lines = format_block(stmt.body, indent + 1, keep_trailing_comments=keep_trailing_comments)
        return lines + [header] + body_lines + [f"{prefix}}}"] + trailing_lines(prefix, trailing)

    if isinstance(stmt, For):
        init = format_for_part(stmt.init)
        cond = format_for_part(stmt.cond)
        inc = format_for_part(stmt.inc)
        header = f"{prefix}for ({init}; {cond}; {inc}) {{"
        body_lines = format_block(stmt.body, indent + 1, keep_trailing_comments=keep_trailing_comments)
        return lines + [header] + body_lines + [f"{prefix}}}"] + trailing_lines(prefix, trailing)
    
    if isinstance(stmt, ForEach):
        header = f"{prefix}for {stmt.name} in {format_expr(stmt.iterable)} {{"
        body_lines = format_block(stmt.body, indent + 1, keep_trailing_comments=keep_trailing_comments)
        return lines + [header] + body_lines + [f"{prefix}}}"] + trailing_lines(prefix, trailing)

    if isinstance(stmt, Block):
        return lines + [f"{prefix}{{"] + format_block(stmt, indent + 1, keep_trailing_comments=keep_trailing_comments) + [f"{prefix}}}"]

    if isinstance(stmt, Comment):
        return lines + [f"{prefix}{stmt.text.rstrip()}"]

    if isinstance(stmt, CheckpointStmt):
        lines.append(f"{prefix}checkpoint {format_expr(stmt.label)}")
        return attach_trailing(lines, prefix, trailing, keep_trailing_comments)

    if isinstance(stmt, Yield):
        if stmt.expr is None:
            lines.append(f"{prefix}yield")
            return attach_trailing(lines, prefix, trailing, keep_trailing_comments)
        lines.append(f"{prefix}yield {format_expr(stmt.expr)}")
        return attach_trailing(lines, prefix, trailing, keep_trailing_comments)

    if isinstance(stmt, Throw):
        lines.append(f"{prefix}throw {format_expr(stmt.expr)}")
        return attach_trailing(lines, prefix, trailing, keep_trailing_comments)

    if isinstance(stmt, TryCatch):
        try_header = f"{prefix}try {{"
        try_lines = format_block(stmt.try_block, indent + 1, keep_trailing_comments=keep_trailing_comments)
        catch_header = f"{prefix}}} catch {stmt.catch_var} {{"
        catch_lines = format_block(stmt.catch_block, indent + 1, keep_trailing_comments=keep_trailing_comments)
        if stmt.finally_block is not None:
            finally_header = f"{prefix}}} finally {{"
            finally_lines = format_block(stmt.finally_block, indent + 1, keep_trailing_comments=keep_trailing_comments)
            out = [try_header] + try_lines + [catch_header] + catch_lines + [finally_header] + finally_lines + [f"{prefix}}}"]
        else:
            out = [try_header] + try_lines + [catch_header] + catch_lines + [f"{prefix}}}"]
        return lines + out + trailing_lines(prefix, trailing)

    if isinstance(stmt, DestructureLet):
        pat = format_pattern(stmt.pattern)
        lines.append(f"{prefix}let {pat} = {format_expr(stmt.expr)}")
        return attach_trailing(lines, prefix, trailing, keep_trailing_comments)

    raise TypeError(f"Unknown stmt node: {stmt!r}")


def format_block(block: Block, indent: int, keep_trailing_comments: bool = False) -> list[str]:
    lines: list[str] = []
    for s in block.stmts:
        lines.extend(format_stmt(s, indent=indent, keep_trailing_comments=keep_trailing_comments))
    return lines


def attach_trailing(lines: list[str], prefix: str, trailing, keep_trailing_comments: bool) -> list[str]:
    if not trailing:
        return lines
    if keep_trailing_comments and lines:
        return lines[:-1] + [lines[-1] + " " + " ".join(t.strip() for t in trailing)]
    return lines + trailing_lines(prefix, trailing)


def trailing_lines(prefix: str, trailing) -> list[str]:
    if not trailing:
        return []
    return [f"{prefix}{text.rstrip()}" for text in trailing]


def format_for_part(part) -> str:
    if part is None:
        return ""
    if isinstance(part, Let):
        return f"let {part.name} = {format_expr(part.expr)}"
    if isinstance(part, ExprStmt):
        return format_expr(part.expr)
    return format_expr(part)


def format_pattern(pattern) -> str:
    if isinstance(pattern, VarPattern):
        return pattern.name
    if isinstance(pattern, ListPattern):
        items = ", ".join(format_pattern(e) for e in pattern.elements)
        return f"[{items}]"
    if isinstance(pattern, RecordPattern):
        pairs = ", ".join(f"{k}: {format_pattern(v)}" for k, v in pattern.fields)
        return f"{{{pairs}}}"
    raise TypeError(f"Unknown pattern node: {pattern!r}")


def format_expr(expr, parent_prec: int = 0) -> str:
    if isinstance(expr, Num):
        return format_number(expr)
    if isinstance(expr, ActionStmt):
        text = f"action {expr.kind}"
        if expr.target is not None:
            text += f" {format_string(expr.target)}"
        if expr.kind in {"tool", "agent", "emit"}:
            text += f" with {format_expr(expr.payload if expr.payload is not None else MapLit([]))}"
            return text
        if expr.kind == "memory_put":
            text += f" {format_expr(expr.payload)}"
            return text
        return text
    if isinstance(expr, Bool):
        return "true" if expr.v else "false"
    if isinstance(expr, Str):
        return format_string(expr.v)
    if isinstance(expr, Nil):
        return "nil"
    if isinstance(expr, Var):
        return expr.name
    if isinstance(expr, Assign):
        text = f"{expr.name} = {format_expr(expr.expr, 1)}"
        return maybe_paren(text, 1, parent_prec)
    if isinstance(expr, Unary):
        inner = format_expr(expr.expr, 7)
        if expr.op == "-" and isinstance(expr.expr, Unary) and expr.expr.op == "-":
            inner = f" {inner}"
        text = f"{expr.op}{inner}"
        return maybe_paren(text, 7, parent_prec)
    if isinstance(expr, Bin):
        prec = bin_prec(expr.op)
        left = format_expr(expr.a, prec)
        right = format_expr(expr.b, prec + 1)
        text = f"{left} {expr.op} {right}"
        return maybe_paren(text, prec, parent_prec)
    if isinstance(expr, Call):
        callee = format_expr(expr.callee, 8)
        args = ", ".join(format_expr(arg) for arg in expr.args)
        return f"{callee}({args})"
    if isinstance(expr, Attr):
        obj = format_expr(expr.obj, 8)
        return f"{obj}.{expr.name}"
    if isinstance(expr, Index):
        seq = format_expr(expr.seq, 8)
        return f"{seq}[{format_expr(expr.index)}]"
    if isinstance(expr, IndexAssign):
        seq = format_expr(expr.seq, 8)
        idx = format_expr(expr.index)
        val = format_expr(expr.value, 1)
        text = f"{seq}[{idx}] = {val}"
        return maybe_paren(text, 1, parent_prec)
    if isinstance(expr, ListLit):
        items = ", ".join(format_expr(item) for item in expr.items)
        return f"[{items}]"
    if isinstance(expr, MapLit):
        pairs = ", ".join(f"{format_expr(k)}: {format_expr(v)}" for k, v in expr.items)
        return f"{{{pairs}}}"
    if isinstance(expr, FnExpr):
        param_text = ", ".join(format_param(param) for param in expr.params)
        return_text = f" -> {expr.return_type}" if expr.return_type else ""
        header = f"fn({param_text}){return_text}"
        if not expr.body.stmts:
            return f"{header} {{}}"
        if len(expr.body.stmts) == 1:
            body_lines = format_stmt(expr.body.stmts[0], indent=0)
            if len(body_lines) == 1:
                return f"{header} {{ {body_lines[0].strip()} }}"
        body_lines = format_block(expr.body, indent=1)
        return f"{header} {{\n" + "\n".join(body_lines) + "\n}"
    if isinstance(expr, FieldAssign):
        obj = format_expr(expr.obj, 8)
        val = format_expr(expr.value, 1)
        text = f"{obj}.{expr.name} = {val}"
        return maybe_paren(text, 1, parent_prec)
    if isinstance(expr, RecordLiteral):
        pairs = ", ".join(f"{k}: {format_expr(v)}" for k, v in expr.fields)
        return f"record {{{pairs}}}"
    # Nodes below are statement-level only and are handled by format_stmt(),
    # not format_expr().  They should never appear as sub-expressions.
    raise TypeError(f"Unknown expr node: {expr!r}")


def format_param(param: Param) -> str:
    if param.type_hint is None:
        return param.name
    return f"{param.name}: {param.type_hint}"


def maybe_paren(text: str, prec: int, parent_prec: int) -> str:
    if prec < parent_prec:
        return f"({text})"
    return text


def bin_prec(op: str) -> int:
    if op in {"||"}:
        return 2
    if op in {"&&"}:
        return 3
    if op in {"==", "!=", "<", ">", "<=", ">="}:
        return 4
    if op in {"+", "-"}:
        return 5
    if op in {"*", "/"}:
        return 6
    return 7


def format_number(num: Num) -> str:
    if num.raw is not None:
        return num.raw
    return str(num.v)


def format_string(value: str) -> str:
    escaped = (
        value.replace("\\", "\\\\")
        .replace("\n", "\\n")
        .replace("\t", "\\t")
        .replace('"', '\\"')
    )
    return f'"{escaped}"'



"""Workflow state helpers."""

from __future__ import annotations


def clone_state(value):
    if isinstance(value, dict):
        return {k: clone_state(v) for k, v in value.items()}
    if isinstance(value, list):
        return [clone_state(v) for v in value]
    return value


def checkpoint_public(entry: dict) -> dict:
    if not isinstance(entry, dict):
        return {}
    out = {}
    for key in ("label", "step", "timestamp", "task_id"):
        if key in entry:
            out[key] = entry[key]
    return out


def checkpoints_public(entries: list) -> list[dict]:
    if not isinstance(entries, list):
        return []
    return [checkpoint_public(entry) for entry in entries if isinstance(entry, dict)]



import io
import os
import tempfile
import unittest
from contextlib import redirect_stdout, redirect_stderr

import nodus as lang
from nodus.tooling.runner import run_source


def run_program(src: str, source_path: str | None = None) -> list[str]:
    result, _vm = run_source(src, filename=source_path)
    return result.get("stdout", "").splitlines()


class PackageTests(unittest.TestCase):
    def test_init_creates_manifest_and_modules_dir(self):
        with tempfile.TemporaryDirectory() as td:
            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "init", "--project-root", td])
            self.assertEqual(exit_code, 0)
            self.assertTrue(os.path.isfile(os.path.join(td, "nodus.toml")))
            self.assertTrue(os.path.isdir(os.path.join(td, ".nodus", "modules")))

    def test_install_creates_dependency_and_lockfile(self):
        with tempfile.TemporaryDirectory() as td:
            utils_dir = os.path.join(td, "utils")
            os.makedirs(utils_dir, exist_ok=True)
            with open(os.path.join(utils_dir, "nodus.toml"), "w", encoding="utf-8") as f:
                f.write('name = "utils"\n')
                f.write('version = "1.2.3"\n')
            with open(os.path.join(utils_dir, "strings.nd"), "w", encoding="utf-8") as f:
                f.write('export fn upper(x) { return x }\n')

            manifest_path = os.path.join(td, "nodus.toml")
            with open(manifest_path, "w", encoding="utf-8") as f:
                f.write('name = "demo"\n')
                f.write('version = "0.1.0"\n\n')
                f.write("[dependencies]\n")
                f.write('utils = { path = "./utils" }\n')

            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "install", "--project-root", td])

            self.assertEqual(exit_code, 0)
            self.assertTrue(os.path.isdir(os.path.join(td, ".nodus", "modules", "utils")))
            self.assertTrue(os.path.isfile(os.path.join(td, "nodus.lock")))
            with open(os.path.join(td, "nodus.lock"), "r", encoding="utf-8") as f:
                lock_text = f.read()
            self.assertIn('[[package]]', lock_text)
            self.assertIn('name = "utils"', lock_text)
            self.assertIn('source = "path"', lock_text)
            self.assertIn('path = "utils"', lock_text)
            self.assertIn('hash = "sha256:', lock_text)

    def test_dependency_import_resolution(self):
        with tempfile.TemporaryDirectory() as td:
            with open(os.path.join(td, "nodus.toml"), "w", encoding="utf-8") as f:
                f.write('name = "demo"\n')
                f.write('version = "0.1.0"\n\n')
            dep_dir = os.path.join(td, ".nodus", "modules", "utils")
            os.makedirs(dep_dir, exist_ok=True)
            with open(os.path.join(dep_dir, "strings.nd"), "w", encoding="utf-8") as f:
                f.write('export fn upper(value) { return "FROM_DEP:" + value }\n')
            main_path = os.path.join(td, "main.nd")
            with open(main_path, "w", encoding="utf-8") as f:
                f.write('import "utils:strings" as s\nprint(s.upper("ok"))\n')

            with open(main_path, "r", encoding="utf-8") as f:
                src = f.read()
            self.assertEqual(run_program(src, source_path=main_path), ["FROM_DEP:ok"])

    def test_package_list_lists_installed_dependencies(self):
        with tempfile.TemporaryDirectory() as td:
            with open(os.path.join(td, "nodus.toml"), "w", encoding="utf-8") as f:
                f.write('name = "demo"\n')
                f.write('version = "0.1.0"\n\n')
                f.write("[dependencies]\n")
                f.write('utils = { path = "./utils" }\n')
            with open(os.path.join(td, "nodus.lock"), "w", encoding="utf-8") as f:
                f.write("[[package]]\n")
                f.write('name = "utils"\n')
                f.write('version = "0.0.0"\n')
                f.write('source = "path"\n')
                f.write('path = "./utils"\n')
                f.write('hash = "sha256:abc123"\n')

            buf = io.StringIO()
            err = io.StringIO()
            with redirect_stdout(buf), redirect_stderr(err):
                exit_code = lang.main(["nodus", "package-list", "--project-root", td])
            self.assertEqual(exit_code, 0)
            self.assertEqual(err.getvalue(), "")
            self.assertIn("utils: path:./utils", buf.getvalue())


if __name__ == "__main__":
    unittest.main()



import io
import tempfile
import unittest
from contextlib import redirect_stdout

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader


def run_program(src: str, source_path: str | None = None) -> list[str]:
    _loader = ModuleLoader(project_root=None)
    code, functions, code_locs = _loader.compile_only(
        src,
        module_name=source_path or "<memory>",
    )
    vm = lang.VM(code, functions, code_locs=code_locs, source_path=source_path)
    buf = io.StringIO()
    with redirect_stdout(buf):
        vm.run()
    return buf.getvalue().splitlines()


class RuntimeStatsTests(unittest.TestCase):
    def test_task_tracking_counts(self):
        src = """
fn worker() {
    yield
}

let a = coroutine(worker)
let b = coroutine(worker)
spawn(a)
spawn(b)
print(len(runtime_tasks()))
run_loop()
let stats = runtime_scheduler_stats()
print(stats["completed"])
"""
        self.assertEqual(run_program(src, source_path="main.nd"), ["2.0", "2.0"])

    def test_resume_count_increments(self):
        src = """
fn worker() {
    yield
    yield
}

let c = coroutine(worker)
spawn(c)
let tasks = runtime_tasks()
let id = tasks[0]["id"]
run_loop()
let info = runtime_task(id)
print(info["resumes"])
"""
        self.assertEqual(run_program(src, source_path="main.nd"), ["3.0"])

    def test_scheduler_sleeping_counts(self):
        src = """
fn sleeper() {
    sleep(50)
}

fn inspector() {
    yield
    let stats = runtime_scheduler_stats()
    print(stats["sleeping"])
}

spawn(coroutine(sleeper))
spawn(coroutine(inspector))
run_loop()
"""
        self.assertEqual(run_program(src, source_path="main.nd"), ["1.0"])

    def test_cli_trace_scheduler(self):
        with tempfile.TemporaryDirectory() as td:
            script = f"{td}/trace.nd"
            with open(script, "w", encoding="utf-8") as f:
                f.write(
                    "fn worker() { yield }\n"
                    "spawn(coroutine(worker))\n"
                    "run_loop()\n"
                )
            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "run", script, "--trace-scheduler"])
            output = buf.getvalue()
            self.assertEqual(exit_code, 0)
            self.assertIn("spawn coroutine #", output)
            self.assertIn("resume coroutine #", output)


if __name__ == "__main__":
    unittest.main()



import io
import os
import tempfile
import unittest
from contextlib import redirect_stdout

import nodus as lang


class CliAllowedPathsTests(unittest.TestCase):
    def test_cli_run_allow_paths_flag(self):
        with tempfile.TemporaryDirectory() as td:
            script = os.path.join(td, "read.nd")
            data = os.path.join(td, "data.txt")
            with open(data, "w", encoding="utf-8") as handle:
                handle.write("ok")
            data_literal = data.replace("\\", "\\\\")
            with open(script, "w", encoding="utf-8") as handle:
                handle.write(f'print(read_file("{data_literal}"))\n')
            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "run", script, "--allow-paths", td])
            self.assertEqual(exit_code, 0)
            self.assertIn("ok", buf.getvalue())

    def test_cli_run_allow_paths_env(self):
        with tempfile.TemporaryDirectory() as td:
            script = os.path.join(td, "read.nd")
            data = os.path.join(td, "data.txt")
            with open(data, "w", encoding="utf-8") as handle:
                handle.write("ok")
            data_literal = data.replace("\\", "\\\\")
            with open(script, "w", encoding="utf-8") as handle:
                handle.write(f'print(read_file("{data_literal}"))\n')
            buf = io.StringIO()
            os.environ["NODUS_ALLOWED_PATHS"] = td
            try:
                with redirect_stdout(buf):
                    exit_code = lang.main(["nodus", "run", script])
            finally:
                os.environ.pop("NODUS_ALLOWED_PATHS", None)
            self.assertEqual(exit_code, 0)
            self.assertIn("ok", buf.getvalue())


if __name__ == "__main__":
    unittest.main()



import unittest

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader


class ScopeTests(unittest.TestCase):
    def test_block_scope_does_not_leak(self):
        src = """
if (true) {
    let x = 1
}
print(x)
"""
        with self.assertRaises(lang.LangSyntaxError) as cm:
            _loader = ModuleLoader(project_root=None)
            _loader.compile_only(src, module_name="main.nd")
        self.assertIn("Undefined variable: x", str(cm.exception))


if __name__ == "__main__":
    unittest.main()



import json
import os
import time
import unittest
import http.client

from nodus.services.server import run_in_thread


class ApiEndpointTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.server, cls.thread = run_in_thread("127.0.0.1", 0, allowed_paths=["."])
        cls.port = cls.server.server_address[1]
        time.sleep(0.05)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server.server_close()

    def request(self, method: str, path: str, payload: dict | None = None):
        conn = http.client.HTTPConnection("127.0.0.1", self.port, timeout=5)
        body = None
        headers = {}
        if payload is not None:
            body = json.dumps(payload)
            headers["Content-Type"] = "application/json"
        conn.request(method, path, body=body, headers=headers)
        resp = conn.getresponse()
        data = resp.read().decode("utf-8")
        conn.close()
        return resp.status, json.loads(data) if data else {}

    def test_disassemble_endpoint(self):
        status, payload = self.request("POST", "/disassemble", {"code": "print(1)", "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertIn("CALL print", payload["disassembly"])

    def test_dis_endpoint(self):
        status, payload = self.request("POST", "/dis", {"code": "print(1)", "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertIn("dis", payload)

    def test_plan_graph_endpoint(self):
        code = (
            "let A = task(fn() { return 1 }, nil)\n"
            "let plan = plan_graph([A])\n"
        )
        status, payload = self.request("POST", "/plan_graph", {"code": code, "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertIsNotNone(payload.get("plan", {}).get("graph_id"))
        self.assertIsNotNone(payload.get("graph_id"))

    def test_graph_plan_endpoint(self):
        code = (
            "let A = task(fn() { return 1 }, nil)\n"
            "let plan = plan_graph([A])\n"
        )
        status, payload = self.request("POST", "/graph/plan", {"code": code, "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertIsNotNone(payload.get("graph_id"))

    def test_graph_run_and_events(self):
        code = (
            "let A = task(fn() { return 1 }, nil)\n"
            "let result = run_graph([A])\n"
        )
        status, payload = self.request("POST", "/graph/run", {"code": code, "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertIsNotNone(payload.get("graph_id"))
        events_status, events = self.request("GET", "/runtime/events")
        self.assertEqual(events_status, 200)
        types = [event["type"] for event in events.get("events", [])]
        self.assertIn("task_start", types)
        self.assertIn("task_success", types)

    def test_graph_resume_endpoint(self):
        code = (
            "let A = task(fn() { return 2 }, nil)\n"
            "let B = task(fn(x) { return x + 1 }, A)\n"
            "let plan = plan_graph([A, B])\n"
        )
        status, payload = self.request("POST", "/graph/plan", {"code": code, "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        graph_id = payload.get("graph_id")
        self.assertIsNotNone(graph_id)
        path = os.path.join(".nodus", "graphs", f"{graph_id}.json")
        os.makedirs(os.path.dirname(path), exist_ok=True)
        state = {
            "graph_id": graph_id,
            "status": "running",
            "tasks": {
                "task_1": {"state": "completed", "result": 2, "attempts": 1},
                "task_2": {"state": "pending", "attempts": 0},
            },
        }
        with open(path, "w", encoding="utf-8") as f:
            json.dump(state, f)
        status, payload = self.request("POST", "/graph/resume", {"graph_id": graph_id})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertIn("task_2", payload.get("result", {}).get("tasks", {}))


if __name__ == "__main__":
    unittest.main()



"""Builtin registry definitions."""

from dataclasses import dataclass


@dataclass
class BuiltinInfo:
    name: str
    arity: int | tuple[int, ...]
    fn: object


BUILTIN_NAMES = {
    "clock",
    "type",
    "runtime_typeof",
    "runtime_fn_name",
    "runtime_fn_arity",
    "runtime_fn_module",
    "runtime_fields",
    "runtime_has",
    "runtime_module_fields",
    "runtime_stack_depth",
    "runtime_stack_frame",
    "runtime_tasks",
    "runtime_task",
    "runtime_scheduler_stats",
    "runtime_time",
    "runtime_events",
    "runtime_clear_events",
    "runtime_event_count",
    "coroutine",
    "resume",
    "coroutine_status",
    "spawn",
    "run_loop",
    "sleep",
    "__sleep",
    "channel",
    "send",
    "recv",
    "close",
    "task",
    "graph",
    "run_graph",
    "plan_graph",
    "resume_graph",
    "run_workflow",
    "plan_workflow",
    "resume_workflow",
    "run_goal",
    "plan_goal",
    "resume_goal",
    "workflow_state",
    "workflow_checkpoints",
    "current_workflow_id",
    "emit",
    "tool_call",
    "tool_available",
    "tool_describe",
    "memory_get",
    "memory_put",
    "memory_delete",
    "memory_keys",
    "recall",
    "remember",
    "suggest",
    "record_outcome",
    "__memory_recall",
    "__memory_remember",
    "__memory_suggest",
    "__memory_record_outcome",
    "__memory_recall_from",
    "__memory_recall_all",
    "__memory_share",
    "agent_call",
    "agent_available",
    "agent_describe",
    "__action_tool",
    "__action_agent",
    "__action_memory_put",
    "__action_memory_get",
    "__action_emit",
    "str",
    "len",
    "collection_len",
    "str_upper",
    "str_lower",
    "str_trim",
    "str_split",
    "str_contains",
    "print",
    "input",
    "keys",
    "values",
    "list_push",
    "list_pop",
    "json_parse",
    "json_stringify",
    "math_abs",
    "math_min",
    "math_max",
    "math_floor",
    "math_ceil",
    "math_sqrt",
    "math_random",
    "read_file",
    "write_file",
    "exists",
    "path_exists",
    "append_file",
    "mkdir",
    "list_dir",
    "path_join",
    "path_dirname",
    "path_basename",
    "path_ext",
    "path_stem",
}






import unittest
import os

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader
from nodus.runtime.module import NodusModule
from nodus.vm.vm import Record


def run_vm(src: str, source_path: str | None = None) -> lang.VM:
    abs_path = os.path.abspath(source_path) if source_path else None
    vm = lang.VM([], {}, code_locs=[], source_path=abs_path)
    _loader = ModuleLoader(project_root=None, vm=vm)
    _loader.load_module_from_source(src, module_name=abs_path or "<memory>")
    return vm


def get_global(vm: lang.VM, name: str):
    if name in vm.globals:
        return vm.globals[name]
    qualified = [value for key, value in vm.globals.items() if key.endswith(f"__{name}")]
    if len(qualified) == 1:
        return qualified[0]
    raise KeyError(name)


class RuntimeReflectionTests(unittest.TestCase):
    def test_function_reflection(self):
        vm = run_vm(
            """
import "std:runtime" as runtime

fn add(a, b) {
    return a + b
}

let fn_name = runtime.fn_name(add)
let fn_arity = runtime.fn_arity(add)
let fn_module = runtime.fn_module(add)
""",
            source_path="main.nd",
        )
        self.assertEqual(get_global(vm, "fn_name"), "add")
        self.assertEqual(get_global(vm, "fn_arity"), 2.0)
        self.assertEqual(get_global(vm, "fn_module"), os.path.abspath("main.nd"))

    def test_record_reflection(self):
        vm = run_vm(
            """
import "std:runtime" as runtime

let user = record { name: "Alice", age: 30 }
let names = runtime.fields(user)
let has_name = runtime.has(user, "name")
let has_email = runtime.has(user, "email")
""",
            source_path="main.nd",
        )
        self.assertEqual(get_global(vm, "names"), ["name", "age"])
        self.assertIs(get_global(vm, "has_name"), True)
        self.assertIs(get_global(vm, "has_email"), False)

    def test_module_reflection(self):
        vm = run_vm(
            """
import "std:runtime" as runtime
import "std:math" as math

let fields = runtime.module_fields(math)
let has_sqrt = runtime.has(math, "sqrt")
let module_type = runtime.typeof(math)
""",
            source_path="main.nd",
        )
        self.assertIn("sqrt", get_global(vm, "fields"))
        self.assertIn("abs", get_global(vm, "fields"))
        self.assertIs(get_global(vm, "has_sqrt"), True)
        self.assertEqual(get_global(vm, "module_type"), "module")
        self.assertIsInstance(vm.globals["math"], (Record, NodusModule))
        self.assertEqual(vm.globals["math"].kind, "module")

    def test_stack_reflection(self):
        vm = run_vm(
            """
import "std:runtime" as runtime

fn inner() {
    return runtime.stack_frame(0)
}

fn outer() {
    return inner()
}

let depth = runtime.stack_depth()
let frame = outer()
""",
            source_path="main.nd",
        )
        self.assertEqual(get_global(vm, "depth"), 0.0)
        frame = get_global(vm, "frame")
        self.assertIsInstance(frame, Record)
        self.assertEqual(frame.fields["name"], "inner")
        self.assertEqual(frame.fields["module"], os.path.abspath("main.nd"))
        self.assertEqual(frame.fields["path"], os.path.abspath("main.nd"))
        self.assertIsInstance(frame.fields["line"], float)
        self.assertIsInstance(frame.fields["column"], float)

    def test_typeof_reflection(self):
        vm = run_vm(
            """
import "std:runtime" as runtime

let int_type = runtime.typeof(10)
let float_type = runtime.typeof(2.5)
let string_type = runtime.typeof("hello")
let bool_type = runtime.typeof(true)
let list_type = runtime.typeof([1, 2])
let record_type = runtime.typeof(record { ok: true })
fn add(a, b) { return a + b }
let function_type = runtime.typeof(add)
let nil_type = runtime.typeof(nil)
""",
            source_path="main.nd",
        )
        self.assertEqual(get_global(vm, "int_type"), "int")
        self.assertEqual(get_global(vm, "float_type"), "float")
        self.assertEqual(get_global(vm, "string_type"), "string")
        self.assertEqual(get_global(vm, "bool_type"), "bool")
        self.assertEqual(get_global(vm, "list_type"), "list")
        self.assertEqual(get_global(vm, "record_type"), "record")
        self.assertEqual(get_global(vm, "function_type"), "function")
        self.assertEqual(get_global(vm, "nil_type"), "nil")


if __name__ == "__main__":
    unittest.main()



import os

from nodus.tooling.runner import run_source


def _write(tmp_path, name: str, content: str) -> str:
    path = tmp_path / name
    path.write_text(content, encoding="utf-8")
    return str(path)


def test_module_globals_isolated(tmp_path):
    mod_a = _write(
        tmp_path,
        "a.nd",
        "export let value = 1\n",
    )
    mod_b = _write(
        tmp_path,
        "b.nd",
        "export let value = 2\n",
    )
    main_path = _write(
        tmp_path,
        "main.nd",
        f'import "{os.path.basename(mod_a)}" as a\nimport "{os.path.basename(mod_b)}" as b\nprint(a.value)\nprint(b.value)\n',
    )
    code = open(main_path, "r", encoding="utf-8").read()
    result, _vm = run_source(code, filename=main_path)
    assert result["ok"] is True
    assert result["stdout"].strip() == "1.0\n2.0"


def test_named_imports_are_live_bindings(tmp_path):
    shared = _write(
        tmp_path,
        "shared.nd",
        "export let value = 1\n",
    )
    main_path = _write(
        tmp_path,
        "main.nd",
        f'import {{ value }} from "{os.path.basename(shared)}"\n'
        "print(value)\n"
        "value = value + 1\n"
        "print(value)\n",
    )
    code = open(main_path, "r", encoding="utf-8").read()
    result, _vm = run_source(code, filename=main_path)
    assert result["ok"] is True
    assert result["stdout"].strip() == "1.0\n2.0"


def test_alias_module_accesses_live_exports(tmp_path):
    shared = _write(
        tmp_path,
        "counter.nd",
        "export let value = 1\n",
    )
    main_path = _write(
        tmp_path,
        "main.nd",
        f'import "{os.path.basename(shared)}" as counter\n'
        "print(counter.value)\n"
        "counter.value = counter.value + 1\n"
        "print(counter.value)\n",
    )
    code = open(main_path, "r", encoding="utf-8").read()
    result, _vm = run_source(code, filename=main_path)
    assert result["ok"] is True
    assert result["stdout"].strip() == "1.0\n2.0"



import http.client
import json
import time
import unittest

from nodus.services.server import run_in_thread


class SessionTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.server, cls.thread = run_in_thread("127.0.0.1", 0, session_timeout_ms=50, allowed_paths=["."])
        cls.port = cls.server.server_address[1]
        time.sleep(0.05)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server.server_close()

    def request(self, method: str, path: str, payload: dict | None = None):
        conn = http.client.HTTPConnection("127.0.0.1", self.port, timeout=5)
        body = None
        headers = {}
        if payload is not None:
            body = json.dumps(payload)
            headers["Content-Type"] = "application/json"
        conn.request(method, path, body=body, headers=headers)
        resp = conn.getresponse()
        data = resp.read().decode("utf-8")
        conn.close()
        return resp.status, json.loads(data) if data else {}

    def test_session_persistence(self):
        status, payload = self.request("POST", "/session")
        self.assertEqual(status, 200)
        session_id = payload["session"]
        self.request("POST", "/execute", {"session": session_id, "code": "x = 10", "filename": "inline.nd"})
        status, payload = self.request("POST", "/execute", {"session": session_id, "code": "print(x)", "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertEqual(payload["stdout"], "10.0\n")

    def test_isolation(self):
        status, payload = self.request("POST", "/session")
        session_a = payload["session"]
        status, payload = self.request("POST", "/session")
        session_b = payload["session"]
        self.request("POST", "/execute", {"session": session_a, "code": "x = 1", "filename": "inline.nd"})
        self.request("POST", "/execute", {"session": session_b, "code": "x = 2", "filename": "inline.nd"})
        _status, payload = self.request("POST", "/execute", {"session": session_a, "code": "print(x)", "filename": "inline.nd"})
        self.assertEqual(payload["stdout"], "1.0\n")
        _status, payload = self.request("POST", "/execute", {"session": session_b, "code": "print(x)", "filename": "inline.nd"})
        self.assertEqual(payload["stdout"], "2.0\n")

    def test_expiration(self):
        status, payload = self.request("POST", "/session")
        session_id = payload["session"]
        time.sleep(0.1)
        status, payload = self.request("GET", "/sessions")
        sessions = {item["id"] for item in payload["sessions"]}
        self.assertNotIn(session_id, sessions)


if __name__ == "__main__":
    unittest.main()



import io
import json
import time
import http.client
import unittest
from contextlib import redirect_stdout

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader

from nodus.services.agent_runtime import AGENT_REGISTRY, register_agent, unregister_agent
from nodus.services.memory_runtime import GLOBAL_MEMORY_STORE
from nodus.services.server import run_in_thread


def run_program(src: str, *, source_path: str = "main.nd"):
    vm = lang.VM([], {}, code_locs=[], source_path=source_path)
    _loader = ModuleLoader(project_root=None, vm=vm)
    buf = io.StringIO()
    with redirect_stdout(buf):
        _loader.load_module_from_source(src, module_name=source_path)
    return vm, buf.getvalue().splitlines()


class AgentRuntimeTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.server, cls.thread = run_in_thread("127.0.0.1", 0, allowed_paths=["."])
        cls.port = cls.server.server_address[1]
        time.sleep(0.05)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server.server_close()

    def setUp(self):
        AGENT_REGISTRY.clear()
        GLOBAL_MEMORY_STORE.load_snapshot({})

    def tearDown(self):
        AGENT_REGISTRY.clear()
        GLOBAL_MEMORY_STORE.load_snapshot({})

    def request(self, method: str, path: str, payload: dict | None = None):
        conn = http.client.HTTPConnection("127.0.0.1", self.port, timeout=5)
        headers = {}
        body = None
        if payload is not None:
            headers["Content-Type"] = "application/json"
            body = json.dumps(payload)
        conn.request(method, path, body=body, headers=headers)
        response = conn.getresponse()
        data = response.read().decode("utf-8")
        conn.close()
        return response.status, json.loads(data) if data else {}

    def test_tool_dispatch_builtin(self):
        _vm, out = run_program(
            """
print(json_stringify(tool_call("nodus_check", {
    "code": "print(1 + 1)",
    "filename": "inline.nd"
})))
"""
        )
        result = json.loads(out[0])
        self.assertTrue(result["ok"])
        self.assertEqual(result["stage"], "check")

    def test_std_tools_wrappers(self):
        _vm, out = run_program(
            """
import "std:tools" as tools
print(json_stringify(tools.available()))
print(json_stringify(tools.describe("nodus_check")))
print(json_stringify(tools.execute("nodus_execute", {
    "code": "print(2)",
    "filename": "inline.nd"
})))
"""
        )
        names = json.loads(out[0])
        info = json.loads(out[1])
        result = json.loads(out[2])
        self.assertIn("nodus_check", names)
        self.assertEqual(info["name"], "nodus_check")
        self.assertTrue(result["ok"])
        self.assertEqual(result["stdout"], "2.0\n")

    def test_memory_store(self):
        _vm, out = run_program(
            """
import "std:memory" as memory
memory.put("topic", "AI SEO")
print(json_stringify(memory.get("topic")))
print(json_stringify(memory.has("topic")))
print(json_stringify(memory.keys()))
print(json_stringify(memory.delete("topic")))
print(json_stringify(memory.get("topic")))
"""
        )
        self.assertEqual(json.loads(out[0]), "AI SEO")
        self.assertTrue(json.loads(out[1]))
        self.assertEqual(json.loads(out[2]), ["topic"])
        self.assertTrue(json.loads(out[3]))
        self.assertIsNone(json.loads(out[4]))

    def test_agent_call_fallback(self):
        _vm, out = run_program(
            """
print(json_stringify(agent_call("summarize", {
    "input": "notes"
})))
"""
        )
        result = json.loads(out[0])
        self.assertFalse(result["ok"])
        self.assertEqual(result["error"]["type"], "agent")

    def test_registered_mock_agent(self):
        register_agent("summarize", lambda payload: {"summary": payload["input"]})
        _vm, out = run_program(
            """
import "std:agent" as agent
print(json_stringify(agent.call("summarize", {
    "input": "hello"
})))
"""
        )
        result = json.loads(out[0])
        self.assertTrue(result["ok"])
        self.assertEqual(result["result"]["summary"], "hello")
        unregister_agent("summarize")

    def test_runtime_events_emitted(self):
        register_agent("summarize", lambda payload: {"summary": payload["input"]})
        vm, _out = run_program(
            """
tool_call("nodus_check", {
    "code": "print(1)",
    "filename": "inline.nd"
})
memory_put("topic", "AI SEO")
memory_get("topic")
agent_call("summarize", {
    "input": "AI SEO"
})
agent_call("missing", {
    "input": "AI SEO"
})
"""
        )
        types = [event.type for event in vm.event_bus.events()]
        self.assertIn("tool_call_start", types)
        self.assertIn("tool_call_complete", types)
        self.assertIn("memory_put", types)
        self.assertIn("memory_get", types)
        self.assertIn("agent_call_start", types)
        self.assertIn("agent_call_complete", types)
        self.assertIn("agent_call_fail", types)

    def test_api_endpoints(self):
        register_agent("summarize", lambda payload: {"summary": payload["input"]})
        status, payload = self.request(
            "POST",
            "/tool/call",
            {"name": "nodus_check", "args": {"code": "print(1 + 1)", "filename": "inline.nd"}},
        )
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])

        status, payload = self.request(
            "POST",
            "/agent/call",
            {"name": "summarize", "payload": {"input": "notes"}},
        )
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertEqual(payload["result"]["summary"], "notes")

        status, payload = self.request("POST", "/memory", {"key": "topic", "value": "AI SEO"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])

        status, payload = self.request("GET", "/memory?key=topic")
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertEqual(payload["result"], "AI SEO")

        status, payload = self.request("GET", "/memory")
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertIn("topic", payload["result"])

        status, payload = self.request("DELETE", "/memory/topic")
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertTrue(payload["result"])

    def test_cli_wrappers(self):
        buf = io.StringIO()
        with redirect_stdout(buf):
            exit_code = lang.main(
                ["nodus", "tool-call", "nodus_check", "--json", "{\"code\":\"print(1)\",\"filename\":\"inline.nd\"}"]
            )
        self.assertEqual(exit_code, 0)
        payload = json.loads(buf.getvalue().strip())
        self.assertTrue(payload["ok"])

        buf = io.StringIO()
        with redirect_stdout(buf):
            exit_code = lang.main(["nodus", "memory-put", "topic", "--json", "\"AI SEO\""])
        self.assertEqual(exit_code, 0)

        buf = io.StringIO()
        with redirect_stdout(buf):
            exit_code = lang.main(["nodus", "memory-get", "topic"])
        self.assertEqual(exit_code, 0)
        self.assertEqual(json.loads(buf.getvalue().strip()), "AI SEO")

        buf = io.StringIO()
        with redirect_stdout(buf):
            exit_code = lang.main(["nodus", "memory-keys"])
        self.assertEqual(exit_code, 0)
        self.assertIn("topic", json.loads(buf.getvalue().strip()))


if __name__ == "__main__":
    unittest.main()



import http.client
import json
import os
import shutil
import tempfile
import time
import unittest

from nodus.services.server import run_in_thread


class SnapshotTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.temp_dir = tempfile.mkdtemp()
        cls.snapshot_dir = os.path.join(cls.temp_dir, "snapshots")
        os.environ["NODUS_SNAPSHOT_DIR"] = cls.snapshot_dir
        cls.server, cls.thread = run_in_thread("127.0.0.1", 0, allowed_paths=["."])
        cls.port = cls.server.server_address[1]
        time.sleep(0.05)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server.server_close()
        shutil.rmtree(cls.temp_dir, ignore_errors=True)
        if "NODUS_SNAPSHOT_DIR" in os.environ:
            del os.environ["NODUS_SNAPSHOT_DIR"]

    def request(self, method: str, path: str, payload: dict | None = None):
        conn = http.client.HTTPConnection("127.0.0.1", self.port, timeout=5)
        body = None
        headers = {}
        if payload is not None:
            body = json.dumps(payload)
            headers["Content-Type"] = "application/json"
        conn.request(method, path, body=body, headers=headers)
        resp = conn.getresponse()
        data = resp.read().decode("utf-8")
        conn.close()
        return resp.status, json.loads(data) if data else {}

    def test_snapshot_persistence(self):
        status, payload = self.request("POST", "/session")
        session_id = payload["session"]
        self.request("POST", "/execute", {"session": session_id, "code": "x = 5", "filename": "inline.nd"})
        status, payload = self.request("POST", "/snapshot", {"session": session_id})
        snapshot_id = payload["snapshot"]
        status, payload = self.request("POST", "/restore", {"snapshot": snapshot_id})
        new_session = payload["session"]
        status, payload = self.request("POST", "/execute", {"session": new_session, "code": "print(x)", "filename": "inline.nd"})
        self.assertEqual(payload["stdout"], "5.0\n")

    def test_snapshot_isolation(self):
        status, payload = self.request("POST", "/session")
        session_id = payload["session"]
        self.request("POST", "/execute", {"session": session_id, "code": "x = 9", "filename": "inline.nd"})
        status, payload = self.request("POST", "/snapshot", {"session": session_id})
        snapshot_id = payload["snapshot"]
        status, payload = self.request("POST", "/restore", {"snapshot": snapshot_id})
        new_session = payload["session"]
        self.assertNotEqual(session_id, new_session)

    def test_snapshot_deletion(self):
        status, payload = self.request("POST", "/session")
        session_id = payload["session"]
        status, payload = self.request("POST", "/snapshot", {"session": session_id})
        snapshot_id = payload["snapshot"]
        status, payload = self.request("DELETE", f"/snapshot/{snapshot_id}")
        self.assertTrue(payload["ok"])
        status, payload = self.request("POST", "/restore", {"snapshot": snapshot_id})
        self.assertIn("error", payload)


if __name__ == "__main__":
    unittest.main()



"""Minimal stdio Debug Adapter Protocol server for Nodus."""

from __future__ import annotations

import io
import json
import os
import sys
import threading
from contextlib import redirect_stderr, redirect_stdout
from dataclasses import dataclass
from typing import BinaryIO

from nodus.runtime.debugger import Debugger, DebuggerQuit, PauseState
from nodus.runtime.errors import format_error_payload
from nodus.runtime.module_loader import ModuleLoader
from nodus.vm.vm import VM


def _read_message(stream: BinaryIO) -> dict | None:
    headers: dict[str, str] = {}
    while True:
        line = stream.readline()
        if not line:
            return None
        if line in {b"\r\n", b"\n"}:
            break
        decoded = line.decode("utf-8").strip()
        if ":" not in decoded:
            continue
        key, value = decoded.split(":", 1)
        headers[key.strip().lower()] = value.strip()
    length_text = headers.get("content-length")
    if not length_text:
        return None
    body = stream.read(int(length_text))
    if not body:
        return None
    return json.loads(body.decode("utf-8"))


def _write_message(stream: BinaryIO, payload: dict) -> None:
    body = json.dumps(payload, separators=(",", ":")).encode("utf-8")
    header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
    stream.write(header)
    stream.write(body)
    stream.flush()


def _resolve_program_path(params: dict) -> str:
    program = params.get("program")
    if not isinstance(program, str) or not program.strip():
        raise ValueError("launch requires a program path")
    return os.path.abspath(program)


def _resolve_project_root(params: dict) -> str | None:
    project_root = params.get("projectRoot")
    if project_root is None:
        return None
    if not isinstance(project_root, str) or not os.path.isdir(project_root):
        raise ValueError(f"Invalid projectRoot: {project_root}")
    return os.path.abspath(project_root)


def _collect_import_state(project_root: str | None) -> dict | None:
    if project_root is None:
        return None
    return {
        "loaded": set(),
        "loading": set(),
        "exports": {},
        "modules": {},
        "module_ids": {},
        "project_root": project_root,
    }


@dataclass
class _FrameScope:
    kind: str
    values: list[tuple[str, object]]


class _ProtocolWriter(io.TextIOBase):
    def __init__(self, session: "DebugSession", category: str):
        self.session = session
        self.category = category

    def write(self, text: str) -> int:
        if text:
            self.session.server.send_event("output", {"category": self.category, "output": text})
        return len(text)

    def flush(self) -> None:
        return None


class ProgrammaticDebugger(Debugger):
    def __init__(self, session: "DebugSession"):
        super().__init__(input_fn=lambda _prompt: "", output_fn=lambda _line: None, start_paused=True)
        self.session = session
        self.condition = threading.Condition()
        self.current_vm = None
        self.stop_count = 0
        self.terminated = False
        self.last_reason: str | None = None
        self.last_pause_state: PauseState | None = None
        self.waiting = False

    def _pause(self, vm, frame, reason: str) -> None:
        locals_view = self.session.get_locals(vm)
        state = PauseState(
            reason=reason,
            module=frame.module,
            line=frame.line,
            col=frame.col,
            function=frame.function,
            locals=locals_view,
        )
        self.last_pause = state
        with self.condition:
            self.current_vm = vm
            self.last_pause_state = state
            self.last_reason = reason
            self.stop_count += 1
            self.waiting = True
            self.condition.notify_all()
        self.session.on_paused(reason, state)
        with self.condition:
            while self.waiting and not self.terminated:
                self.condition.wait()
            if self.terminated:
                raise DebuggerQuit()

    def mark_terminated(self) -> None:
        with self.condition:
            self.terminated = True
            self.condition.notify_all()

    def wait_for_stop(self, previous_count: int, timeout: float = 1.0) -> bool:
        with self.condition:
            if self.stop_count > previous_count or self.terminated:
                return True
            self.condition.wait(timeout)
            return self.stop_count > previous_count or self.terminated

    def _resume(self, action: str) -> None:
        with self.condition:
            vm = self.current_vm
            if vm is None:
                return
            if action == "continue":
                self.resume_continue(vm)
            elif action == "stepIn":
                self.resume_step_in(vm)
            elif action == "next":
                self.resume_step_over(vm)
            elif action == "stepOut":
                self.resume_step_out(vm)
            self.waiting = False
            self.condition.notify_all()

    def continue_execution(self) -> None:
        self._resume("continue")

    def step_in_execution(self) -> None:
        self._resume("stepIn")

    def step_over_execution(self) -> None:
        self._resume("next")

    def step_out_execution(self) -> None:
        self._resume("stepOut")

    def pause_execution(self) -> None:
        self.request_pause()

    def terminate_execution(self) -> None:
        self.request_terminate()
        with self.condition:
            self.terminated = True
            self.waiting = False
            self.condition.notify_all()


class DebugSession:
    def __init__(self, server: "DebugAdapterServer"):
        self.server = server
        self.program: str | None = None
        self.project_root: str | None = None
        self.vm: VM | None = None
        self.debugger: ProgrammaticDebugger | None = None
        self.thread: threading.Thread | None = None
        self.breakpoint_lines: dict[str, set[int]] = {}
        self.frame_handles: dict[int, int] = {}
        self.scope_handles: dict[int, _FrameScope] = {}
        self.next_scope_handle = 1
        self._terminated_sent = False

    def launch(self, params: dict) -> None:
        program = _resolve_program_path(params)
        if not os.path.isfile(program):
            raise ValueError(f"File not found: {program}")
        project_root = _resolve_project_root(params)
        with open(program, "r", encoding="utf-8") as handle:
            source = handle.read()
        debugger = ProgrammaticDebugger(self)
        for path, lines in self.breakpoint_lines.items():
            for line in lines:
                debugger.set_breakpoint(path, line)
        vm = VM(
            [],
            {},
            code_locs=[],
            source_path=program,
            debug=True,
            debugger=debugger,
        )
        loader = ModuleLoader(project_root=project_root, vm=vm, debugger=debugger)
        bytecode, functions, code_locs = loader.compile_only(
            source,
            module_name=program,
            base_dir=os.path.dirname(os.path.abspath(program)) if program else os.getcwd(),
        )
        vm.reset_program(bytecode, functions, code_locs=code_locs, source_path=program)
        vm.source_code = source
        self.program = program
        self.project_root = project_root
        self.vm = vm
        self.debugger = debugger
        self.thread = threading.Thread(target=self._run_vm, name="nodus-dap", daemon=True)
        self.thread.start()
        debugger.wait_for_stop(0, timeout=1.0)

    def _run_vm(self) -> None:
        if self.vm is None or self.debugger is None:
            return
        stdout_proxy = _ProtocolWriter(self, "stdout")
        stderr_proxy = _ProtocolWriter(self, "stderr")
        try:
            with redirect_stdout(stdout_proxy), redirect_stderr(stderr_proxy):
                self.vm.run()
        except DebuggerQuit:
            pass
        except Exception as err:
            self.server.send_event("output", {"category": "stderr", "output": f"{format_error_payload({'message': str(err)})}\n"})
        finally:
            self.debugger.mark_terminated()
            self._send_terminated()

    def _send_terminated(self) -> None:
        if self._terminated_sent:
            return
        self._terminated_sent = True
        self.server.send_event("terminated", {})

    def on_paused(self, reason: str, state: PauseState) -> None:
        reason_map = {
            "start": "entry",
            "breakpoint": "breakpoint",
            "pause": "pause",
            "step": "step",
            "next": "step",
            "stepOut": "step",
        }
        body = {
            "reason": reason_map.get(reason, "pause"),
            "threadId": 1,
            "allThreadsStopped": True,
        }
        if state.line is not None:
            body["line"] = state.line
        self.server.send_event("stopped", body)

    def disconnect(self) -> None:
        if self.debugger is not None:
            self.debugger.terminate_execution()
        if self.thread is not None:
            self.thread.join(timeout=1.0)
        self._send_terminated()

    def set_breakpoints(self, source_path: str, lines: list[int]) -> list[dict]:
        normalized = os.path.abspath(source_path)
        previous = self.breakpoint_lines.get(normalized, set())
        self.breakpoint_lines[normalized] = set(lines)
        if self.debugger is not None:
            for line in previous - self.breakpoint_lines[normalized]:
                self.debugger.clear_breakpoint(normalized, line)
            for line in self.breakpoint_lines[normalized] - previous:
                self.debugger.set_breakpoint(normalized, line)
        return [{"verified": True, "line": line} for line in lines]

    def continue_execution(self) -> None:
        if self.debugger is not None:
            self.debugger.continue_execution()

    def pause_execution(self) -> None:
        if self.debugger is not None:
            self.debugger.pause_execution()

    def step_in_execution(self) -> None:
        if self.debugger is not None:
            self.debugger.step_in_execution()

    def step_over_execution(self) -> None:
        if self.debugger is not None:
            self.debugger.step_over_execution()

    def step_out_execution(self) -> None:
        if self.debugger is not None:
            self.debugger.step_out_execution()

    def _collect_frames(self) -> list[dict]:
        if self.vm is None:
            return []
        if not self.vm.frames:
            path, line, col = self.vm.current_loc()
            return [{"name": "main", "path": path or self.program, "line": line or 1, "col": col or 1, "locals": self.vm.globals, "params": []}]
        frames = self.vm.reflection_frames()
        out: list[dict] = []
        for index in range(len(frames)):
            frame = frames[-1 - index]
            fn = frame.closure.function if frame.closure is not None else self.vm.functions.get(frame.fn_name)
            module_path = None
            if fn is not None and 0 <= fn.addr < len(self.vm.code_locs):
                module_path = self.vm.code_locs[fn.addr][0]
            if index == 0 and len(frames) == len(self.vm.frames):
                current_path, current_line, current_col = self.vm.current_loc()
                line = current_line
                col = current_col
                path = current_path or module_path or self.program
            else:
                line = frame.call_line
                col = frame.call_col
                path = frame.call_path or module_path or self.program
            # Merge dict-based locals and slot-indexed locals_array for full view
            merged_locals: dict = dict(frame.locals)
            if frame.locals_array is not None and frame.locals_name_to_slot is not None:
                for var_name, slot in frame.locals_name_to_slot.items():
                    arr_val = frame.locals_array[slot]
                    if hasattr(arr_val, "value"):
                        arr_val = arr_val.value
                    merged_locals[var_name] = arr_val
            out.append(
                {
                    "name": self.vm.display_name(frame.fn_name),
                    "path": path or self.program,
                    "line": line or 1,
                    "col": col or 1,
                    "locals": merged_locals,
                    "params": list(fn.params) if fn is not None else [],
                }
            )
        return out

    def stack_trace(self) -> list[dict]:
        frames = self._collect_frames()
        self.frame_handles = {}
        out: list[dict] = []
        for index, frame in enumerate(frames, start=1):
            self.frame_handles[index] = index - 1
            out.append(
                {
                    "id": index,
                    "name": frame["name"],
                    "source": {"path": frame["path"], "name": os.path.basename(frame["path"])},
                    "line": frame["line"],
                    "column": frame["col"],
                }
            )
        return out

    def scopes(self, frame_id: int) -> list[dict]:
        frame_index = self.frame_handles.get(frame_id)
        frames = self._collect_frames()
        if frame_index is None or frame_index >= len(frames):
            raise ValueError(f"Unknown frame id: {frame_id}")
        frame = frames[frame_index]
        values = frame["locals"] or {}
        params = set(frame["params"])
        arguments = [(name, values[name]) for name in frame["params"] if name in values]
        locals_ = [(name, value) for name, value in sorted(values.items()) if name not in params]
        scopes: list[dict] = []
        if arguments:
            handle = self._store_scope("Arguments", arguments)
            scopes.append({"name": "Arguments", "presentationHint": "arguments", "variablesReference": handle, "expensive": False})
        handle = self._store_scope("Locals", locals_)
        scopes.append({"name": "Locals", "presentationHint": "locals", "variablesReference": handle, "expensive": False})
        return scopes

    def _store_scope(self, kind: str, values: list[tuple[str, object]]) -> int:
        handle = self.next_scope_handle
        self.next_scope_handle += 1
        self.scope_handles[handle] = _FrameScope(kind=kind, values=values)
        return handle

    def variables(self, variables_reference: int) -> list[dict]:
        scope = self.scope_handles.get(variables_reference)
        if scope is None:
            raise ValueError(f"Unknown variables reference: {variables_reference}")
        return [
            {
                "name": name,
                "value": self.value_to_string(value),
                "variablesReference": 0,
            }
            for name, value in scope.values
        ]

    def get_locals(self, vm: VM) -> dict:
        values = vm.current_locals()
        if values is None:
            values = vm.globals
        out: dict[str, object] = {}
        for name, value in (values or {}).items():
            out[name] = getattr(value, "value", value)
        return out

    def value_to_string(self, value: object) -> str:
        if self.vm is None:
            return str(value)
        return self.vm.value_to_string(getattr(value, "value", value), quote_strings=True)


class DebugAdapterServer:
    def __init__(self, input_stream: BinaryIO | None = None, output_stream: BinaryIO | None = None):
        self.input_stream = input_stream if input_stream is not None else sys.stdin.buffer
        self.output_stream = output_stream if output_stream is not None else sys.stdout.buffer
        self.output_lock = threading.Lock()
        self.exit_code = 0
        self.seq = 1
        self.session = DebugSession(self)

    def run(self) -> int:
        while True:
            message = _read_message(self.input_stream)
            if message is None:
                break
            if self.handle_message(message):
                break
        return self.exit_code

    def handle_message(self, message: dict) -> bool:
        if message.get("type") != "request":
            return False
        command = message.get("command")
        request_seq = message.get("seq")
        arguments = message.get("arguments", {})
        try:
            if command == "initialize":
                self.send_response(
                    request_seq,
                    command,
                    {
                        "supportsConfigurationDoneRequest": False,
                        "supportsPauseRequest": True,
                        "supportsStepInTargetsRequest": False,
                    },
                )
                self.send_event("initialized", {})
                return False
            if command == "launch":
                self.session.launch(arguments)
                self.send_response(request_seq, command, {})
                return False
            if command == "disconnect":
                self.session.disconnect()
                self.send_response(request_seq, command, {})
                return True
            if command == "setBreakpoints":
                source = arguments.get("source", {})
                source_path = source.get("path")
                if not isinstance(source_path, str) or not source_path:
                    raise ValueError("setBreakpoints requires source.path")
                if "lines" in arguments:
                    lines = [int(line) for line in arguments.get("lines", [])]
                else:
                    lines = [int(entry.get("line")) for entry in arguments.get("breakpoints", []) if entry.get("line") is not None]
                body = {"breakpoints": self.session.set_breakpoints(source_path, lines)}
                self.send_response(request_seq, command, body)
                return False
            if command == "continue":
                self.session.continue_execution()
                self.send_response(request_seq, command, {"allThreadsContinued": True})
                return False
            if command == "pause":
                self.session.pause_execution()
                self.send_response(request_seq, command, {})
                return False
            if command == "next":
                self.session.step_over_execution()
                self.send_response(request_seq, command, {})
                return False
            if command == "stepIn":
                self.session.step_in_execution()
                self.send_response(request_seq, command, {})
                return False
            if command == "stepOut":
                self.session.step_out_execution()
                self.send_response(request_seq, command, {})
                return False
            if command == "stackTrace":
                frames = self.session.stack_trace()
                self.send_response(request_seq, command, {"stackFrames": frames, "totalFrames": len(frames)})
                return False
            if command == "scopes":
                frame_id = int(arguments.get("frameId"))
                self.send_response(request_seq, command, {"scopes": self.session.scopes(frame_id)})
                return False
            if command == "variables":
                variables_reference = int(arguments.get("variablesReference"))
                self.send_response(request_seq, command, {"variables": self.session.variables(variables_reference)})
                return False
            self.send_error(request_seq, command, f"Unsupported command: {command}")
            return False
        except Exception as err:
            self.send_error(request_seq, command, str(err))
            return command == "disconnect"

    def _send(self, payload: dict) -> None:
        with self.output_lock:
            payload = {"seq": self.seq, **payload}
            self.seq += 1
            _write_message(self.output_stream, payload)

    def send_response(self, request_seq: int, command: str, body: dict) -> None:
        self._send(
            {
                "type": "response",
                "request_seq": request_seq,
                "success": True,
                "command": command,
                "body": body,
            }
        )

    def send_error(self, request_seq: int, command: str | None, message: str) -> None:
        self._send(
            {
                "type": "response",
                "request_seq": request_seq,
                "success": False,
                "command": command,
                "message": message,
            }
        )

    def send_event(self, event: str, body: dict) -> None:
        self._send({"type": "event", "event": event, "body": body})


def run_stdio_server() -> int:
    return DebugAdapterServer().run()


if __name__ == "__main__":
    raise SystemExit(run_stdio_server())



"""Runtime stats helpers for scheduler and coroutine introspection."""

from __future__ import annotations

import time

_START = time.monotonic()


def runtime_time_ms() -> float:
    return (time.monotonic() - _START) * 1000.0


def task_snapshot(coroutine) -> dict:
    return {
        "id": float(coroutine.id) if coroutine.id is not None else None,
        "name": coroutine.name,
        "module": coroutine.module,
        "status": coroutine.state,
        "resumes": float(coroutine.resume_count),
        "created_time": float(coroutine.created_time) if coroutine.created_time is not None else None,
        "last_resume": float(coroutine.last_resume) if coroutine.last_resume is not None else None,
        "last_run_time": float(coroutine.last_run_time) if coroutine.last_run_time is not None else None,
    }


def scheduler_stats(scheduler) -> dict:
    return {
        "ready": float(len(scheduler.ready_queue)),
        "sleeping": float(len(scheduler.sleeping_tasks)),
        "completed": float(len(scheduler.completed_tasks)),
        "spawned": float(scheduler.total_tasks_spawned),
        "resumes": float(scheduler.total_resumes),
        "ready_queue": [float(task.id) for task in scheduler.ready_queue if task.id is not None],
        "sleeping_tasks": [float(task_id) for task_id in sorted(scheduler.sleeping_tasks)],
        "completed_tasks": [float(task.id) for task in scheduler.completed_tasks if task.id is not None],
    }



"""Runtime module loader for Nodus."""

from __future__ import annotations

import os
from dataclasses import dataclass

from nodus.builtins.nodus_builtins import BUILTIN_NAMES, BuiltinInfo
from nodus.compiler.compiler import Compiler, wrap_bytecode
from nodus.frontend.lexer import Tok, tokenize
from nodus.frontend.parser import Parser
from nodus.frontend.ast.ast_nodes import (
    Assign,
    Attr,
    Bin,
    Block,
    Call,
    Comment,
    ExportFrom,
    ExportList,
    ExprStmt,
    FnDef,
    GoalDef,
    If,
    Import,
    Index,
    IndexAssign,
    Let,
    ListLit,
    MapLit,
    ModuleInfo,
    WorkflowDef,
    Unary,
    While,
    For,
)
from nodus.runtime.diagnostics import LangRuntimeError, LangSyntaxError
from nodus.runtime.bytecode_cache import load_cached_bytecode, write_cached_bytecode
from nodus.runtime.dependency_graph import DependencyGraph
from nodus.runtime.module import LiveBinding, ModuleBytecode, NodusModule
from nodus.tooling.project import NODUS_DIRNAME, MODULES_DIRNAME, find_project_root
from nodus.vm.vm import VM


@dataclass
class ImportSpec:
    path: str
    names: list[str]
    alias: str | None
    resolved_path: str


@dataclass
class ExportFromSpec:
    path: str
    names: list[str]
    resolved_path: str


@dataclass
class ParsedModule:
    module_id: str
    source: str
    ast: list
    module_info: ModuleInfo
    imports: list[Import]
    export_from: list[ExportFrom]
    base_dir: str


@dataclass
class ModuleMetadata:
    module_id: str
    exports: set[str]
    import_names: set[str]
    import_specs: list[ImportSpec]
    export_from_specs: list[ExportFromSpec]
    module_info: ModuleInfo
    parsed: ParsedModule | None


class ModuleLoader:
    def __init__(
        self,
        *,
        project_root: str | None = None,
        host_globals: dict | None = None,
        host_builtins: dict[str, BuiltinInfo] | None = None,
        extra_builtins: set[str] | None = None,
        vm: VM | None = None,
        debugger=None,
    ) -> None:
        self.project_root = project_root
        self.host_globals = host_globals or {}
        self.host_builtins = host_builtins or {}
        self.extra_builtins = set(extra_builtins or [])
        self._modules: dict[str, NodusModule] = {}
        self._metadata: dict[str, ModuleMetadata] = {}
        self._parsed: dict[str, ParsedModule] = {}
        self._loading: set[str] = set()
        self._import_state: dict = {
            "loaded": set(),
            "loading": set(),
            "exports": {},
            "modules": {},
            "module_ids": {},
            "project_root": project_root,
        }
        self._dependency_graph: DependencyGraph | None = DependencyGraph.load(project_root)
        self._recompiled_modules: set[str] = set()
        self._vm = vm
        self._debugger = debugger

    def resolve_import(self, import_path: str, base_dir: str, tok: Tok | None, module_id: str) -> str:
        if "project_root" not in self._import_state:
            self._import_state["project_root"] = self.project_root
        return resolve_import_path(
            import_path,
            base_dir,
            self._import_state,
            tok,
            module_id,
        )

    def load_module_from_path(self, path: str, *, initial_globals: dict | None = None) -> NodusModule:
        module_id = os.path.abspath(path)
        base_dir = os.path.dirname(module_id)
        return self._load_module(module_id, base_dir=base_dir, source_path=module_id, initial_globals=initial_globals)

    def load_module_from_source(
        self,
        source: str,
        *,
        module_name: str = "<memory>",
        base_dir: str | None = None,
        initial_globals: dict | None = None,
    ) -> NodusModule:
        module_id = module_name
        base_dir = base_dir or os.getcwd()
        source_path = None
        if module_name not in {"<memory>"} and os.path.isfile(module_name):
            source_path = os.path.abspath(module_name)
        return self._load_module(
            module_id,
            base_dir=base_dir,
            source=source,
            source_path=source_path,
            initial_globals=initial_globals,
        )

    def compile_only(self, source: str, *, module_name: str, base_dir: str | None = None) -> tuple[dict, dict, list]:
        base_dir = base_dir or os.getcwd()
        module_id = module_name
        metadata = self._build_metadata(module_id, base_dir=base_dir, source=source)
        bytecode, functions, code_locs = self._compile_module(metadata)
        return bytecode, functions, code_locs

    def _load_module(
        self,
        module_id: str,
        *,
        base_dir: str,
        source: str | None = None,
        source_path: str | None = None,
        initial_globals: dict | None = None,
    ) -> NodusModule:
        if module_id in self._modules:
            return self._modules[module_id]
        if module_id in self._loading:
            raise LangRuntimeError("import", f"Cyclic import detected: {module_id}", path=module_id)

        self._loading.add(module_id)
        try:
            metadata = self._build_metadata(module_id, base_dir=base_dir, source=source, source_path=source_path)
            bytecode_unit = self._load_or_compile_module_bytecode(metadata, source_path=source_path)
            module = NodusModule(
                name=os.path.basename(module_id) if module_id not in {"<memory>"} else module_id,
                path=module_id,
                bytecode=bytecode_unit.code,
                functions=bytecode_unit.functions,
                code_locs=bytecode_unit.code_locs,
                bytecode_unit=bytecode_unit,
                globals={},
                exports={},
                host_globals=self.host_globals,
                host_builtins=self.host_builtins,
                initialized=False,
            )
            self._modules[module_id] = module

            import_bindings, dep_modules = self._resolve_import_bindings(metadata)
            module.globals.update(import_bindings)
            if initial_globals:
                module.globals.update(initial_globals)
            self._execute_module(module, source_path=source_path)
            module.exports = self._build_exports(metadata, module, dep_modules)
            module.initialized = True
            return module
        finally:
            self._loading.discard(module_id)

    def _load_or_compile_module_bytecode(
        self,
        metadata: ModuleMetadata,
        *,
        source_path: str | None,
    ) -> ModuleBytecode:
        can_reuse_cache = source_path is not None and self._can_skip_reprocessing(source_path)
        if can_reuse_cache and source_path is not None:
            cached = load_cached_bytecode(self.project_root, source_path)
            if cached is not None:
                if not cached.module_metadata:
                    cached.module_metadata = self._serialize_module_metadata(metadata)
                    write_cached_bytecode(self.project_root, source_path, cached)
                return cached
        bytecode, functions, code_locs = self._compile_module(metadata)
        bytecode_unit = ModuleBytecode(
            code=bytecode,
            functions=functions,
            constants=list(bytecode.get("constants", [])),
            code_locs=code_locs,
            symbol_table={
                "defs": sorted(metadata.module_info.defs),
                "exports": sorted(metadata.exports),
                "imports": sorted(metadata.import_names),
            },
            module_metadata=self._serialize_module_metadata(metadata),
        )
        if source_path is not None:
            write_cached_bytecode(self.project_root, source_path, bytecode_unit)
            self._record_dependency_graph(metadata, source_path)
            self._recompiled_modules.add(os.path.abspath(source_path))
        return bytecode_unit

    def _execute_module(self, module: NodusModule, *, source_path: str | None) -> None:
        vm = self._vm
        if vm is None:
            vm = VM(
                module.bytecode,
                module.functions,
                code_locs=module.code_locs,
                module_globals=module.globals,
                host_globals=self.host_globals,
                source_path=source_path,
            )
            self._vm = vm
        else:
            vm.reset_program(
                module.bytecode,
                module.functions,
                code_locs=module.code_locs,
                source_path=source_path,
                module_globals=module.globals,
                host_globals=self.host_globals,
            )
        if self.host_builtins:
            vm.builtins.update(self.host_builtins)
        if self._debugger is not None:
            vm.debugger = self._debugger
            vm.debug = True
        vm.run()

    def _ensure_dependency_graph(self) -> DependencyGraph | None:
        if self.project_root is None:
            return None
        if self._dependency_graph is None:
            self._dependency_graph = DependencyGraph.load(self.project_root)
        return self._dependency_graph

    def _record_dependency_graph(self, metadata: ModuleMetadata, source_path: str) -> None:
        graph = self._ensure_dependency_graph()
        if graph is None:
            return
        graph.update_module(
            source_path,
            [spec.resolved_path for spec in metadata.import_specs] + [spec.resolved_path for spec in metadata.export_from_specs],
            os.stat(source_path).st_mtime_ns,
        )
        graph.save()

    def _can_skip_reprocessing(self, source_path: str) -> bool:
        graph = self._ensure_dependency_graph()
        if graph is None:
            return False
        return not self._is_module_stale(os.path.abspath(source_path), seen=set())

    def _is_module_stale(self, module_path: str, *, seen: set[str]) -> bool:
        normalized = os.path.abspath(module_path)
        if normalized in seen:
            return False
        if normalized in self._recompiled_modules:
            return True
        seen.add(normalized)
        graph = self._ensure_dependency_graph()
        if graph is None:
            return True
        node = graph.get(normalized)
        if node is None or not os.path.isfile(normalized):
            return True
        current_mtime = os.stat(normalized).st_mtime_ns
        if current_mtime != node.last_compiled_mtime:
            return True
        for dependency in node.imported_modules:
            dep_path = os.path.abspath(dependency)
            if dep_path in self._recompiled_modules:
                return True
            if self._is_module_stale(dep_path, seen=seen.copy()):
                return True
        return False

    def _serialize_module_metadata(self, metadata: ModuleMetadata) -> dict[str, object]:
        return {
            "module_id": metadata.module_id,
            "exports": sorted(metadata.exports),
            "import_names": sorted(metadata.import_names),
            "import_specs": [
                {
                    "path": spec.path,
                    "names": list(spec.names),
                    "alias": spec.alias,
                    "resolved_path": spec.resolved_path,
                }
                for spec in metadata.import_specs
            ],
            "export_from_specs": [
                {
                    "path": spec.path,
                    "names": list(spec.names),
                    "resolved_path": spec.resolved_path,
                }
                for spec in metadata.export_from_specs
            ],
            "module_info": {
                "defs": sorted(metadata.module_info.defs),
                "exports": sorted(metadata.module_info.exports),
                "explicit_exports": metadata.module_info.explicit_exports,
            },
        }

    def _build_metadata_from_cached_bytecode(self, module_id: str, bytecode_unit: ModuleBytecode) -> ModuleMetadata | None:
        payload = bytecode_unit.module_metadata
        if not isinstance(payload, dict):
            return None
        raw_import_specs = payload.get("import_specs", [])
        raw_export_from_specs = payload.get("export_from_specs", [])
        raw_module_info = payload.get("module_info", {})
        if not isinstance(raw_import_specs, list) or not isinstance(raw_export_from_specs, list) or not isinstance(raw_module_info, dict):
            return None
        import_specs: list[ImportSpec] = []
        for item in raw_import_specs:
            if not isinstance(item, dict):
                return None
            import_specs.append(
                ImportSpec(
                    path=str(item.get("path", "")),
                    names=[str(name) for name in item.get("names", []) if isinstance(name, str)],
                    alias=str(item["alias"]) if item.get("alias") is not None else None,
                    resolved_path=str(item.get("resolved_path", "")),
                )
            )
        export_from_specs: list[ExportFromSpec] = []
        for item in raw_export_from_specs:
            if not isinstance(item, dict):
                return None
            export_from_specs.append(
                ExportFromSpec(
                    path=str(item.get("path", "")),
                    names=[str(name) for name in item.get("names", []) if isinstance(name, str)],
                    resolved_path=str(item.get("resolved_path", "")),
                )
            )
        module_info = ModuleInfo(
            path=module_id,
            defs=set(str(name) for name in raw_module_info.get("defs", []) if isinstance(name, str)),
            exports=set(str(name) for name in raw_module_info.get("exports", []) if isinstance(name, str)),
            imports={},
            aliases={},
            explicit_exports=bool(raw_module_info.get("explicit_exports", False)),
            qualified={},
        )
        return ModuleMetadata(
            module_id=module_id,
            exports=set(str(name) for name in payload.get("exports", []) if isinstance(name, str)),
            import_names=set(str(name) for name in payload.get("import_names", []) if isinstance(name, str)),
            import_specs=import_specs,
            export_from_specs=export_from_specs,
            module_info=module_info,
            parsed=None,
        )

    def _build_metadata(
        self,
        module_id: str,
        *,
        base_dir: str,
        source: str | None = None,
        source_path: str | None = None,
    ) -> ModuleMetadata:
        if module_id in self._metadata:
            return self._metadata[module_id]

        if "project_root" not in self._import_state or self._import_state["project_root"] is None:
            ensure_project_root(self._import_state, base_dir, source_path)
            self.project_root = self._import_state.get("project_root")
        self._ensure_dependency_graph()

        if source_path is not None and self._can_skip_reprocessing(source_path):
            cached = load_cached_bytecode(self.project_root, source_path)
            if cached is not None:
                cached_metadata = self._build_metadata_from_cached_bytecode(module_id, cached)
                if cached_metadata is not None:
                    self._metadata[module_id] = cached_metadata
                    return cached_metadata

        parsed = self._parse_module(module_id, base_dir=base_dir, source=source, source_path=source_path)
        import_specs: list[ImportSpec] = []
        export_from_specs: list[ExportFromSpec] = []
        import_names: set[str] = set()

        for stmt in parsed.imports:
            tok = getattr(stmt, "_tok", None)
            resolved = self.resolve_import(stmt.path, parsed.base_dir, tok, parsed.module_id)
            import_specs.append(ImportSpec(path=stmt.path, names=list(stmt.names or []), alias=stmt.alias, resolved_path=resolved))

        for stmt in parsed.export_from:
            tok = getattr(stmt, "_tok", None)
            resolved = self.resolve_import(stmt.path, parsed.base_dir, tok, parsed.module_id)
            export_from_specs.append(ExportFromSpec(path=stmt.path, names=list(stmt.names or []), resolved_path=resolved))

        for stmt, spec in zip(parsed.imports, import_specs):
            dep_meta = self._build_metadata(spec.resolved_path, base_dir=os.path.dirname(spec.resolved_path), source_path=spec.resolved_path)
            if spec.names:
                missing = [name for name in spec.names if name not in dep_meta.exports]
                if missing:
                    tok = getattr(stmt, "_tok", None)
                    line = tok.line if tok is not None else None
                    col = tok.col if tok is not None else None
                    raise LangRuntimeError(
                        "import",
                        f"Import failed: {spec.resolved_path} does not export {', '.join(missing)}",
                        line=line,
                        col=col,
                        path=spec.resolved_path,
                    )
                import_names.update(spec.names)
            elif spec.alias:
                import_names.add(spec.alias)
            else:
                import_names.update(dep_meta.exports)

        for stmt, spec in zip(parsed.export_from, export_from_specs):
            dep_meta = self._build_metadata(spec.resolved_path, base_dir=os.path.dirname(spec.resolved_path), source_path=spec.resolved_path)
            missing = [name for name in spec.names if name not in dep_meta.exports]
            if missing:
                tok = getattr(stmt, "_tok", None)
                line = tok.line if tok is not None else None
                col = tok.col if tok is not None else None
                raise LangRuntimeError(
                    "import",
                    f"Re-export failed: {spec.resolved_path} does not export {', '.join(missing)}",
                    line=line,
                    col=col,
                    path=spec.resolved_path,
                )

        metadata = ModuleMetadata(
            module_id=module_id,
            exports=set(parsed.module_info.exports),
            import_names=import_names,
            import_specs=import_specs,
            export_from_specs=export_from_specs,
            module_info=parsed.module_info,
            parsed=parsed,
        )
        self._metadata[module_id] = metadata
        return metadata

    def _parse_module(
        self,
        module_id: str,
        *,
        base_dir: str,
        source: str | None = None,
        source_path: str | None = None,
    ) -> ParsedModule:
        if module_id in self._parsed:
            return self._parsed[module_id]
        if source is None:
            with open(module_id, "r", encoding="utf-8") as handle:
                source = handle.read()
        toks = tokenize(source)
        ast = Parser(toks).parse()
        set_module_on_tree(ast, module_id)
        module_info = collect_module_info(ast, module_id, "")
        imports = [stmt for stmt in ast if isinstance(stmt, Import)]
        export_from = [stmt for stmt in ast if isinstance(stmt, ExportFrom)]
        parsed = ParsedModule(
            module_id=module_id,
            source=source,
            ast=ast,
            module_info=module_info,
            imports=imports,
            export_from=export_from,
            base_dir=base_dir,
        )
        self._parsed[module_id] = parsed
        return parsed

    def _compile_module(self, metadata: ModuleMetadata) -> tuple[dict, dict, list]:
        if metadata.parsed is None:
            raise LangRuntimeError("compile", f"Module metadata for {metadata.module_id} is not available for compilation", path=metadata.module_id)
        module_info = metadata.module_info
        module_info.imports = {name: name for name in metadata.import_names}
        module_info.qualified = {name: name for name in module_info.defs}
        builtin_names = set(BUILTIN_NAMES)
        if self.extra_builtins:
            builtin_names.update(self.extra_builtins)
        compiler = Compiler(module_infos={metadata.module_id: module_info}, module_defs_index={}, builtin_names=builtin_names)
        code, functions, code_locs = compiler.compile_program(metadata.parsed.ast)
        bytecode = wrap_bytecode(
            code,
            module_name=metadata.module_id,
            exports=sorted(metadata.exports),
        )
        return bytecode, functions, code_locs

    def _resolve_import_bindings(self, metadata: ModuleMetadata) -> tuple[dict[str, object], dict[str, NodusModule]]:
        bindings: dict[str, object] = {}
        modules: dict[str, NodusModule] = {}
        for spec in metadata.import_specs:
            module = self._load_module(spec.resolved_path, base_dir=os.path.dirname(spec.resolved_path), source_path=spec.resolved_path)
            modules[spec.resolved_path] = module
            if spec.names:
                for name in spec.names:
                    bindings[name] = module.export_binding(name)
            elif spec.alias:
                bindings[spec.alias] = module
            else:
                for name in module.exports:
                    bindings[name] = module.export_binding(name)
        return bindings, modules

    def _build_exports(
        self,
        metadata: ModuleMetadata,
        module: NodusModule,
        dep_modules: dict[str, NodusModule],
    ) -> dict[str, object]:
        exports: dict[str, object] = {}
        for name in metadata.exports:
            if name in module.globals or name in metadata.module_info.defs:
                exports[name] = LiveBinding(module, name)
                continue
            resolved = None
            for spec in metadata.export_from_specs:
                if name in spec.names:
                    dep = dep_modules.get(spec.resolved_path)
                    if dep is None:
                        dep = self._load_module(spec.resolved_path, base_dir=os.path.dirname(spec.resolved_path), source_path=spec.resolved_path)
                    resolved = dep.export_binding(name)
                    break
            if resolved is not None:
                exports[name] = resolved
        return exports


def set_module_on_tree(node, module_id: str):
    if node is None:
        return
    if isinstance(node, list):
        for item in node:
            set_module_on_tree(item, module_id)
        return
    if not hasattr(node, "__dict__"):
        return
    setattr(node, "_module", module_id)
    for key, value in node.__dict__.items():
        if key in {"_tok", "_module"}:
            continue
        if isinstance(value, Tok):
            continue
        if isinstance(value, list):
            for item in value:
                set_module_on_tree(item, module_id)
        else:
            set_module_on_tree(value, module_id)


def ensure_project_root(import_state: dict, base_dir: str, source_path: str | None):
    if "project_root" not in import_state:
        import_state["project_root"] = None
    if import_state["project_root"] is None:
        env_root = os.environ.get("NODUS_PROJECT_ROOT")
        if env_root:
            import_state["project_root"] = env_root

    project_root = import_state.get("project_root")
    if project_root is None:
        discovered_root = find_project_root(base_dir)
        import_state["project_root"] = discovered_root or base_dir
        return

    project_root = os.path.abspath(project_root)
    if not os.path.isdir(project_root):
        raise LangRuntimeError(
            "import",
            f"Invalid project root: {project_root}",
            path=source_path,
        )
    import_state["project_root"] = project_root


def try_resolve_with_extensions(base_path: str) -> str | None:
    if base_path.endswith(".nd") or base_path.endswith(".tl"):
        full = os.path.abspath(base_path)
        if os.path.exists(full):
            return full
        return None

    candidates = [
        os.path.abspath(base_path + ".nd"),
        os.path.abspath(base_path + ".tl"),
        os.path.abspath(os.path.join(base_path, "index.nd")),
        os.path.abspath(os.path.join(base_path, "index.tl")),
    ]
    for candidate in candidates:
        if os.path.exists(candidate):
            return candidate
    return None


def resolve_with_extensions(base_path: str, import_path: str, tok: Tok | None, module_id: str) -> str:
    if base_path.endswith(".nd") or base_path.endswith(".tl"):
        full = os.path.abspath(base_path)
        if os.path.exists(full):
            return full
        import_error(f"Import not found: {import_path} (tried {full})", tok, module_id)

    candidates = [
        os.path.abspath(base_path + ".nd"),
        os.path.abspath(base_path + ".tl"),
        os.path.abspath(os.path.join(base_path, "index.nd")),
        os.path.abspath(os.path.join(base_path, "index.tl")),
    ]

    for candidate in candidates:
        if os.path.exists(candidate):
            return candidate

    import_error(
        f"Import not found: {import_path} (tried {', '.join(candidates)})",
        tok,
        module_id,
    )


def resolve_import_path(
    import_path: str,
    base_dir: str,
    import_state: dict,
    tok: Tok | None,
    module_id: str,
) -> str:
    project_root = os.path.abspath(import_state.get("project_root") or base_dir)
    modules_dir = os.path.join(project_root, NODUS_DIRNAME, MODULES_DIRNAME)

    if ":" in import_path and not import_path.startswith("std:"):
        package_name, package_path = import_path.split(":", 1)
        if not package_name or not package_path:
            import_error("Invalid package import: use package:module", tok, module_id)
        if package_name.startswith(".") or package_name.startswith(("/", "\\")):
            import_error("Invalid package import: package name is invalid", tok, module_id)
        package_base = os.path.normpath(os.path.join(modules_dir, package_name, package_path.replace("/", os.sep).replace("\\", os.sep)))
        package_root = os.path.normpath(os.path.join(modules_dir, package_name))
        if not package_base.startswith(package_root):
            import_error("Invalid package import: path escapes dependency directory", tok, module_id)
        return resolve_with_extensions(package_base, import_path, tok, module_id)

    if import_path.startswith("std:"):
        name = import_path[4:]
        if not name:
            import_error("Invalid std import: missing module name (use std:strings)", tok, module_id)
        if name.startswith(("/", "\\")):
            import_error("Invalid std import: std modules cannot start with '/'", tok, module_id)
        std_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), "..", "stdlib"))
        name = name.replace("/", os.sep).replace("\\", os.sep)
        base = os.path.normpath(os.path.join(std_dir, name))
        std_dir_norm = os.path.normpath(std_dir)
        if not base.startswith(std_dir_norm):
            import_error("Invalid std import: path escapes std directory", tok, module_id)
        return resolve_with_extensions(base, import_path, tok, module_id)

    if os.path.isabs(import_path):
        base = import_path
    elif import_path.startswith("."):
        base = os.path.join(base_dir, import_path)
        base_norm = os.path.normcase(os.path.normpath(base))
        root_norm = os.path.normcase(os.path.normpath(project_root))
        try:
            if os.path.commonpath([base_norm, root_norm]) != root_norm:
                import_error("Invalid import: path escapes project root", tok, module_id)
        except ValueError:
            import_error("Invalid import: path escapes project root", tok, module_id)
    else:
        base = os.path.join(project_root, import_path)

    base = os.path.normpath(base)
    resolved = try_resolve_with_extensions(base)
    if resolved is not None:
        return resolved

    modules_base = os.path.normpath(os.path.join(modules_dir, import_path))
    resolved = try_resolve_with_extensions(modules_base)
    if resolved is not None:
        return resolved

    std_base = _resolve_std_base(import_path, tok, module_id)
    resolved = try_resolve_with_extensions(std_base)
    if resolved is not None:
        return resolved

    return resolve_with_extensions(base, import_path, tok, module_id)


def import_error(message: str, tok: Tok | None, module_id: str):
    line = tok.line if tok is not None else None
    col = tok.col if tok is not None else None
    raise LangRuntimeError("import", message, line=line, col=col, path=module_id)


def collect_module_info(stmts: list, module_id: str, prefix: str) -> ModuleInfo:
    defs: set[str] = set()
    explicit_exports = False
    explicit: set[str] = set()
    reexports: set[str] = set()

    def walk_expr(e):
        if e is None:
            return
        if isinstance(e, Assign):
            defs.add(e.name)
            walk_expr(e.expr)
            return
        if isinstance(e, Unary):
            walk_expr(e.expr)
            return
        if isinstance(e, Bin):
            walk_expr(e.a)
            walk_expr(e.b)
            return
        if isinstance(e, ListLit):
            for item in e.items:
                walk_expr(item)
            return
        if isinstance(e, MapLit):
            for k, v in e.items:
                walk_expr(k)
                walk_expr(v)
            return
        if isinstance(e, Index):
            walk_expr(e.seq)
            walk_expr(e.index)
            return
        if isinstance(e, IndexAssign):
            walk_expr(e.seq)
            walk_expr(e.index)
            walk_expr(e.value)
            return
        if isinstance(e, Attr):
            walk_expr(e.obj)
            return
        if isinstance(e, Call):
            walk_expr(e.callee)
            for arg in e.args:
                walk_expr(arg)
            return

    def walk_stmt(s):
        nonlocal explicit_exports
        if isinstance(s, Comment):
            return
        if isinstance(s, Let):
            defs.add(s.name)
            if s.exported:
                explicit_exports = True
                explicit.add(s.name)
            walk_expr(s.expr)
            return
        if isinstance(s, WorkflowDef):
            defs.add(s.name)
            return
        if isinstance(s, GoalDef):
            defs.add(s.name)
            return
        if isinstance(s, FnDef):
            defs.add(s.name)
            if s.exported:
                explicit_exports = True
                explicit.add(s.name)
            return
        if isinstance(s, ExportList):
            explicit_exports = True
            explicit.update(s.names)
            return
        if isinstance(s, ExportFrom):
            explicit_exports = True
            reexports.update(s.names)
            return
        if isinstance(s, ExprStmt):
            walk_expr(s.expr)
            return
        if isinstance(s, If):
            walk_expr(s.cond)
            walk_stmt(s.then_branch)
            if s.else_branch is not None:
                walk_stmt(s.else_branch)
            return
        if isinstance(s, While):
            walk_expr(s.cond)
            walk_stmt(s.body)
            return
        if isinstance(s, For):
            walk_stmt(s.init)
            walk_expr(s.cond)
            walk_expr(s.inc)
            walk_stmt(s.body)
            return
        if isinstance(s, Block):
            for inner in s.stmts:
                walk_stmt(inner)
            return

    for stmt in stmts:
        walk_stmt(stmt)

    exports = (explicit | reexports) if explicit_exports else set(defs)

    if explicit_exports:
        missing = [name for name in explicit if name not in defs]
        if missing:
            line = None
            col = None
            for stmt in stmts:
                if isinstance(stmt, ExportList):
                    tok = getattr(stmt, "_tok", None)
                    if tok is not None:
                        line = tok.line
                        col = tok.col
                        break
            raise LangSyntaxError(
                f"Exported name(s) not defined in module: {', '.join(missing)}",
                line=line,
                col=col,
                path=module_id,
            )

    qualified = {name: f"{prefix}{name}" for name in defs}

    return ModuleInfo(
        path=module_id,
        defs=defs,
        exports=exports,
        imports={},
        aliases={},
        explicit_exports=explicit_exports,
        qualified=qualified,
    )


def _resolve_std_base(import_path: str, tok: Tok | None, module_id: str) -> str:
    if import_path.startswith(("/", "\\")):
        import_error("Invalid std import: std modules cannot start with '/'", tok, module_id)
    std_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), "..", "stdlib"))
    name = import_path.replace("/", os.sep).replace("\\", os.sep)
    base = os.path.normpath(os.path.join(std_dir, name))
    if not base.startswith(os.path.normpath(std_dir)):
        import_error("Invalid std import: path escapes std directory", tok, module_id)
    return base



"""Sandbox enforcement utilities for runner execution."""

from __future__ import annotations

import io
import time
from contextlib import contextmanager, redirect_stderr, redirect_stdout

from nodus.support.config import EXECUTION_TIMEOUT_MS, MAX_STEPS, MAX_STDOUT_CHARS, MAX_STACK_DEPTH
from nodus.runtime.diagnostics import RuntimeLimitExceeded


class LimitedBuffer(io.StringIO):
    def __init__(self, max_chars: int | None):
        super().__init__()
        self.max_chars = max_chars
        self._count = 0

    def write(self, s: str) -> int:
        if self.max_chars is not None and self._count + len(s) > self.max_chars:
            raise RuntimeLimitExceeded("stdout limit exceeded")
        self._count += len(s)
        return super().write(s)


def configure_vm_limits(
    vm,
    *,
    max_steps: int | None = MAX_STEPS,
    timeout_ms: int | None = EXECUTION_TIMEOUT_MS,
) -> None:
    vm.max_steps = max_steps
    if timeout_ms is None:
        vm.deadline = None
    else:
        vm.deadline = time.monotonic() + (timeout_ms / 1000.0)
    vm.max_frames = MAX_STACK_DEPTH


@contextmanager
def capture_output(*, max_stdout_chars: int | None = MAX_STDOUT_CHARS):
    stdout = LimitedBuffer(max_stdout_chars)
    stderr = io.StringIO()
    with redirect_stdout(stdout), redirect_stderr(stderr):
        yield stdout, stderr



"""Runtime event tracing for the Nodus VM."""

from __future__ import annotations

import json
from dataclasses import dataclass

from nodus.runtime.runtime_stats import runtime_time_ms


def _normalize_value(value):
    if value is None:
        return None
    if isinstance(value, bool):
        return value
    if isinstance(value, int):
        return float(value)
    if isinstance(value, float):
        return float(value)
    if isinstance(value, list):
        return [_normalize_value(item) for item in value]
    if isinstance(value, dict):
        return {key: _normalize_value(val) for key, val in value.items()}
    return value


@dataclass
class RuntimeEvent:
    type: str
    timestamp: float
    coroutine_id: int | None = None
    name: str | None = None
    data: dict | None = None

    def to_dict(self) -> dict:
        return {
            "type": self.type,
            "timestamp": float(self.timestamp),
            "coroutine": float(self.coroutine_id) if self.coroutine_id is not None else None,
            "name": self.name,
            "data": _normalize_value(self.data) if self.data is not None else None,
        }


class RuntimeEventBus:
    def __init__(self, enabled: bool = True):
        self._enabled = enabled
        self._events: list[RuntimeEvent] = []
        self._sinks: list[object] = []

    def emit(self, event: RuntimeEvent) -> None:
        if not self._enabled:
            return
        self._events.append(event)
        for sink in self._sinks:
            sink.emit(event)

    def emit_event(
        self,
        event_type: str,
        *,
        coroutine_id: int | None = None,
        name: str | None = None,
        data: dict | None = None,
    ) -> None:
        self.emit(
            RuntimeEvent(
                event_type,
                runtime_time_ms(),
                coroutine_id=coroutine_id,
                name=name,
                data=data,
            )
        )

    def events(self) -> list[RuntimeEvent]:
        return list(self._events)

    def clear(self) -> None:
        self._events.clear()

    def enable(self) -> None:
        self._enabled = True

    def disable(self) -> None:
        self._enabled = False

    def add_sink(self, sink) -> None:
        self._sinks.append(sink)


def format_event(event: RuntimeEvent) -> str:
    text = f"[{event.timestamp:.1f}ms] {event.type}"
    if event.coroutine_id is not None:
        text += f" #{event.coroutine_id}"
    if event.name:
        text += f" {event.name}"
    if event.data:
        for key, value in event.data.items():
            text += f" {key}={value}"
    return text


class HumanReadableEventSink:
    def __init__(self, write_line):
        self.write_line = write_line

    def emit(self, event: RuntimeEvent) -> None:
        self.write_line(format_event(event))


class JsonEventSink:
    def __init__(self, write_line):
        self.write_line = write_line

    def emit(self, event: RuntimeEvent) -> None:
        self.write_line(json.dumps(event.to_dict(), separators=(",", ":")))



"""AST pretty-printer for Nodus."""

from __future__ import annotations

from dataclasses import is_dataclass

from nodus.frontend.ast.ast_nodes import (
    Assign,
    ActionStmt,
    Attr,
    Bin,
    Block,
    Bool,
    Call,
    Comment,
    ExportFrom,
    ExportList,
    ExprStmt,
    FnDef,
    For,
    GoalDef,
    GoalStep,
    If,
    Import,
    Index,
    IndexAssign,
    Let,
    ListLit,
    MapLit,
    ModuleAlias,
    Nil,
    Num,
    Param,
    Print,
    Return,
    Str,
    Unary,
    Var,
    WorkflowStateDecl,
    CheckpointStmt,
    WorkflowDef,
    WorkflowStep,
    While,
)


def format_ast(stmts: list, compact: bool = False) -> str:
    printer = AstPrinter(compact=compact)
    printer.emit("Module", 0)
    for stmt in stmts:
        printer.visit(stmt, 1)
    return "\n".join(printer.lines)


class AstPrinter:
    def __init__(self, compact: bool = False):
        self.lines: list[str] = []
        self.compact = compact

    def emit(self, text: str, indent: int):
        self.lines.append(("  " * indent) + text)

    def visit(self, node, indent: int):
        if node is None:
            self.emit("<none>", indent)
            return
        if isinstance(node, list):
            for item in node:
                self.visit(item, indent)
            return
        if isinstance(node, Num):
            self.emit(f"Number {self.format_number(node)}", indent)
            return
        if isinstance(node, Bool):
            self.emit(f"Bool {self.format_bool(node.v)}", indent)
            return
        if isinstance(node, Str):
            self.emit(f"String {self.format_string(node.v)}", indent)
            return
        if isinstance(node, Nil):
            self.emit("Nil", indent)
            return
        if isinstance(node, Var):
            self.emit(f"Identifier {node.name}", indent)
            return
        if isinstance(node, Unary):
            self.emit(f"Unary op={node.op}", indent)
            if self.compact:
                self.visit(node.expr, indent + 1)
            else:
                self.emit("expr:", indent + 1)
                self.visit(node.expr, indent + 2)
            return
        if isinstance(node, Bin):
            self.emit(f"Binary op={node.op}", indent)
            if self.compact:
                self.visit(node.a, indent + 1)
                self.visit(node.b, indent + 1)
            else:
                self.emit("left:", indent + 1)
                self.visit(node.a, indent + 2)
                self.emit("right:", indent + 1)
                self.visit(node.b, indent + 2)
            return
        if isinstance(node, Assign):
            self.emit(f"Assign name={node.name}", indent)
            if self.compact:
                self.visit(node.expr, indent + 1)
            else:
                self.emit("value:", indent + 1)
                self.visit(node.expr, indent + 2)
            return
        if isinstance(node, ListLit):
            self.emit("List", indent)
            if self.compact:
                for item in node.items:
                    self.visit(item, indent + 1)
            else:
                self.emit("items:", indent + 1)
                for item in node.items:
                    self.visit(item, indent + 2)
            return
        if isinstance(node, MapLit):
            self.emit("Map", indent)
            if not node.items:
                return
            if self.compact:
                for key, value in node.items:
                    self.emit("entry:", indent + 1)
                    self.visit(key, indent + 2)
                    self.visit(value, indent + 2)
            else:
                self.emit("entries:", indent + 1)
                for key, value in node.items:
                    self.emit("entry:", indent + 2)
                    self.emit("key:", indent + 3)
                    self.visit(key, indent + 4)
                    self.emit("value:", indent + 3)
                    self.visit(value, indent + 4)
            return
        if isinstance(node, Index):
            self.emit("Index", indent)
            if self.compact:
                self.visit(node.seq, indent + 1)
                self.visit(node.index, indent + 1)
            else:
                self.emit("target:", indent + 1)
                self.visit(node.seq, indent + 2)
                self.emit("index:", indent + 1)
                self.visit(node.index, indent + 2)
            return
        if isinstance(node, IndexAssign):
            self.emit("IndexAssign", indent)
            if self.compact:
                self.visit(node.seq, indent + 1)
                self.visit(node.index, indent + 1)
                self.visit(node.value, indent + 1)
            else:
                self.emit("target:", indent + 1)
                self.visit(node.seq, indent + 2)
                self.emit("index:", indent + 1)
                self.visit(node.index, indent + 2)
                self.emit("value:", indent + 1)
                self.visit(node.value, indent + 2)
            return
        if isinstance(node, Attr):
            self.emit(f"Attr name={node.name}", indent)
            if self.compact:
                self.visit(node.obj, indent + 1)
            else:
                self.emit("object:", indent + 1)
                self.visit(node.obj, indent + 2)
            return
        if isinstance(node, Call):
            self.emit("Call", indent)
            if self.compact:
                self.visit(node.callee, indent + 1)
                for arg in node.args:
                    self.visit(arg, indent + 1)
            else:
                self.emit("callee:", indent + 1)
                self.visit(node.callee, indent + 2)
                self.emit("args:", indent + 1)
                for arg in node.args:
                    self.visit(arg, indent + 2)
            return
        if isinstance(node, Let):
            suffix = " exported" if node.exported else ""
            type_part = f" type={node.type_hint}" if node.type_hint else ""
            self.emit(f"Let name={node.name}{type_part}{suffix}", indent)
            if self.compact:
                self.visit(node.expr, indent + 1)
            else:
                self.emit("value:", indent + 1)
                self.visit(node.expr, indent + 2)
            return
        if isinstance(node, Print):
            self.emit("Print", indent)
            if self.compact:
                self.visit(node.expr, indent + 1)
            else:
                self.emit("value:", indent + 1)
                self.visit(node.expr, indent + 2)
            return
        if isinstance(node, ExprStmt):
            self.emit("ExprStmt", indent)
            self.visit(node.expr, indent + 1)
            return
        if isinstance(node, Block):
            self.emit("Block", indent)
            for stmt in node.stmts:
                self.visit(stmt, indent + 1)
            return
        if isinstance(node, Comment):
            self.emit(f"Comment {self.format_string(node.text.rstrip())}", indent)
            return
        if isinstance(node, If):
            self.emit("If", indent)
            if self.compact:
                self.visit(node.cond, indent + 1)
                self.visit(node.then_branch, indent + 1)
            else:
                self.emit("cond:", indent + 1)
                self.visit(node.cond, indent + 2)
                self.emit("then:", indent + 1)
                self.visit(node.then_branch, indent + 2)
            if node.else_branch is not None:
                if self.compact:
                    self.visit(node.else_branch, indent + 1)
                else:
                    self.emit("else:", indent + 1)
                    self.visit(node.else_branch, indent + 2)
            return
        if isinstance(node, While):
            self.emit("While", indent)
            if self.compact:
                self.visit(node.cond, indent + 1)
                self.visit(node.body, indent + 1)
            else:
                self.emit("cond:", indent + 1)
                self.visit(node.cond, indent + 2)
                self.emit("body:", indent + 1)
                self.visit(node.body, indent + 2)
            return
        if isinstance(node, For):
            self.emit("For", indent)
            if self.compact:
                if node.init is None:
                    self.emit("<none>", indent + 1)
                else:
                    self.visit(node.init, indent + 1)
                if node.cond is None:
                    self.emit("<none>", indent + 1)
                else:
                    self.visit(node.cond, indent + 1)
                if node.inc is None:
                    self.emit("<none>", indent + 1)
                else:
                    self.visit(node.inc, indent + 1)
                self.visit(node.body, indent + 1)
            else:
                self.emit("init:", indent + 1)
                if node.init is None:
                    self.emit("<none>", indent + 2)
                else:
                    self.visit(node.init, indent + 2)
                self.emit("cond:", indent + 1)
                if node.cond is None:
                    self.emit("<none>", indent + 2)
                else:
                    self.visit(node.cond, indent + 2)
                self.emit("inc:", indent + 1)
                if node.inc is None:
                    self.emit("<none>", indent + 2)
                else:
                    self.visit(node.inc, indent + 2)
                self.emit("body:", indent + 1)
                self.visit(node.body, indent + 2)
            return
        if isinstance(node, FnDef):
            suffix = " exported" if node.exported else ""
            params = ", ".join(self.format_param(param) for param in node.params)
            return_part = f" return={node.return_type}" if node.return_type else ""
            self.emit(f"FnDef name={node.name} params=[{params}]{return_part}{suffix}", indent)
            if self.compact:
                self.visit(node.body, indent + 1)
            else:
                self.emit("body:", indent + 1)
                self.visit(node.body, indent + 2)
            return
        if isinstance(node, WorkflowDef):
            self.emit(f"WorkflowDef name={node.name}", indent)
            if node.states:
                self.emit("states:", indent + 1)
                for state in node.states:
                    self.visit(state, indent + 2)
            for step in node.steps:
                self.visit(step, indent + 1)
            return
        if isinstance(node, WorkflowStateDecl):
            self.emit(f"WorkflowStateDecl name={node.name}", indent)
            if self.compact:
                self.visit(node.value, indent + 1)
            else:
                self.emit("value:", indent + 1)
                self.visit(node.value, indent + 2)
            return
        if isinstance(node, WorkflowStep):
            dep_part = f" deps=[{', '.join(node.deps)}]" if node.deps else ""
            self.emit(f"WorkflowStep name={node.name}{dep_part}", indent)
            if node.options is not None:
                self.emit("options:", indent + 1)
                self.visit(node.options, indent + 2)
            self.emit("body:", indent + 1)
            self.visit(node.body, indent + 2)
            return
        if isinstance(node, CheckpointStmt):
            self.emit("CheckpointStmt", indent)
            if self.compact:
                self.visit(node.label, indent + 1)
            else:
                self.emit("label:", indent + 1)
                self.visit(node.label, indent + 2)
            return
        if isinstance(node, GoalDef):
            self.emit(f"GoalDef name={node.name}", indent)
            if node.states:
                self.emit("states:", indent + 1)
                for state in node.states:
                    self.visit(state, indent + 2)
            for step in node.steps:
                self.visit(step, indent + 1)
            return
        if isinstance(node, GoalStep):
            dep_part = f" deps=[{', '.join(node.deps)}]" if node.deps else ""
            self.emit(f"GoalStep name={node.name}{dep_part}", indent)
            if node.options is not None:
                self.emit("options:", indent + 1)
                self.visit(node.options, indent + 2)
            self.emit("body:", indent + 1)
            self.visit(node.body, indent + 2)
            return
        if isinstance(node, ActionStmt):
            target = f" target={self.format_string(node.target)}" if node.target is not None else ""
            self.emit(f"ActionStmt kind={node.kind}{target}", indent)
            if node.payload is not None:
                self.emit("payload:", indent + 1)
                self.visit(node.payload, indent + 2)
            return
        if isinstance(node, Return):
            self.emit("Return", indent)
            if node.expr is not None:
                if self.compact:
                    self.visit(node.expr, indent + 1)
                else:
                    self.emit("value:", indent + 1)
                    self.visit(node.expr, indent + 2)
            return
        if isinstance(node, Import):
            line = f"Import path={self.format_string(node.path)}"
            if node.alias:
                line += f" alias={node.alias}"
            if node.names:
                line += f" names=[{', '.join(node.names)}]"
            self.emit(line, indent)
            return
        if isinstance(node, ExportList):
            self.emit(f"ExportList names=[{', '.join(node.names)}]", indent)
            return
        if isinstance(node, ExportFrom):
            self.emit(f"ExportFrom path={self.format_string(node.path)} names=[{', '.join(node.names)}]", indent)
            return
        if isinstance(node, ModuleAlias):
            exports = ", ".join(sorted(node.exports.keys()))
            self.emit(f"ModuleAlias alias={node.alias} exports=[{exports}]", indent)
            return

        if is_dataclass(node):
            self.emit(type(node).__name__, indent)
            return

        self.emit(f"<unknown {type(node).__name__}>", indent)

    def format_number(self, node: Num) -> str:
        if node.raw is not None:
            return node.raw
        return str(node.v)

    def format_bool(self, value: bool) -> str:
        return "true" if value else "false"

    def format_string(self, value: str) -> str:
        escaped = (
            value.replace("\\", "\\\\")
            .replace("\n", "\\n")
            .replace("\t", "\\t")
            .replace('"', '\\"')
        )
        return f"\"{escaped}\""

    def format_param(self, param: Param) -> str:
        if param.type_hint is None:
            return param.name
        return f"{param.name}: {param.type_hint}"



"""Runtime configuration defaults for service mode."""

import os

MAX_STEPS = 10_000_000
MAX_STDOUT_CHARS = 20_000
MAX_STACK_DEPTH = 10_000
EXECUTION_TIMEOUT_MS = 200
SESSION_TIMEOUT_MS = 300_000
MAX_SESSIONS = 100
SNAPSHOT_DIR = os.environ.get("NODUS_SNAPSHOT_DIR", ".nodus/snapshots")
MAX_SNAPSHOTS = 200
SERVER_HOST = "127.0.0.1"
SERVER_PORT = 7331
WORKER_SWEEP_INTERVAL_MS = 500
